import os
import sys
import traceback

from PyQt5.QtCore import Qt, QThread, QSize
from PyQt5.QtWidgets import (
    QDialog, QVBoxLayout, QHBoxLayout, QLabel, QFileDialog
)
from PyQt5 import QtGui

from .. import exception_handler
from .. import myutils, html_utils, workers, widgets, load, apps

class repeatDataPrepWindow(QDialog):
    def __init__(self, parent=None) -> None:
        super().__init__(parent)

        name = 'repeat data prep'
        
        logger, logs_path, log_path, log_filename = myutils.setupLogger(
            module=name
        )

        self.logger = logger
        self.log_path = log_path
        self.log_filename = log_filename
        self.logs_path = logs_path

        self.logger.info(f'Initializing {name}...')

        self.cancel = True

        self.setWindowTitle(f'Cell-ACDC {name}')
        self.funcDescription = f'Cell-ACDC {name}'

        instructions = [
            'Press <b>start<b> button',
            '<b>Select experiment</b> folder or specific Position folder',
            'Select which <b>channels</b> to reapply data prep to',
            '<b>Wait</b> until process ends'
        ]

        txt = html_utils.paragraph(f"""
            This utility is used to <b>re-apply data prep</b> steps such as 
            cropping and aligning.<br><br>
            This is needed when you are adding <b>new channels</b> 
            to already prepped data.<br><br>
            How to use it:
            {html_utils.to_list(instructions, ordered=True)}
        """)

        layout = QVBoxLayout()
        textLayout = QHBoxLayout()

        pixmap = QtGui.QIcon(":cog_play.svg").pixmap(QSize(64,64))
        iconLabel = QLabel()
        iconLabel.setPixmap(pixmap)

        textLayout.addWidget(iconLabel, alignment=Qt.AlignTop)
        textLayout.addSpacing(20)
        textLayout.addWidget(QLabel(txt))
        textLayout.addStretch(1)

        buttonsLayout = QHBoxLayout()
        stopButton = widgets.stopPushButton('Stop process')
        startButton = widgets.playPushButton('    Start     ')
        cancelButton = widgets.cancelPushButton('Close')

        buttonsLayout.addStretch(1)
        buttonsLayout.addWidget(cancelButton)
        buttonsLayout.addSpacing(20)
        buttonsLayout.addWidget(startButton)
        buttonsLayout.addWidget(stopButton)

        self.startButton = startButton
        self.stopButton = stopButton

        progressBarLayout = QHBoxLayout()
        self.progressBar = widgets.QProgressBarWithETA(parent=self)
        progressBarLayout.addWidget(self.progressBar)
        progressBarLayout.addWidget(self.progressBar.ETA_label)
        self.logConsole = widgets.QLogConsole(parent=self)

        layout.addLayout(textLayout)
        layout.addSpacing(20)
        layout.addLayout(buttonsLayout)
        layout.addSpacing(20)
        layout.addLayout(progressBarLayout)
        layout.addWidget(self.logConsole)

        self.setLayout(layout)

        cancelButton.clicked.connect(self.close)
        startButton.clicked.connect(self.start)
        stopButton.clicked.connect(self.stop)
    
    def showEvent(self, event: QtGui.QShowEvent) -> None:
        self.startButton.setFixedWidth(self.stopButton.width())
        self.stopButton.hide()
        return super().showEvent(event)

    @exception_handler    
    def start(self):
        self.startButton.hide()
        self.stopButton.show()

        MostRecentPath = myutils.getMostRecentPath()
        exp_path = QFileDialog.getExistingDirectory(
            self, 'Select experiment folder or specific Position folder', 
            MostRecentPath
        )
        if not exp_path:
            self.logger.info('No path selected. Process stopped.')
            self.stop()
            return
        
        myutils.addToRecentPaths(exp_path, logger=self.logger)
        
        folder_type = myutils.determine_folder_type(exp_path)
        is_pos_folder, is_images_folder, exp_path = folder_type

        if is_pos_folder:
            pos_path = exp_path
            exp_path = os.path.dirname(pos_path)
            posFoldernames = [os.path.basename(pos_path)]
        elif is_images_folder:
            pos_path = os.path.dirname(exp_path)
            exp_path = os.path.dirname(pos_path)
            posFoldernames = [os.path.basename(pos_path)]
        else:
            select_folder = load.select_exp_folder()
            values = select_folder.get_values_dataprep(exp_path)
            if not values:
                self.criticalNotValidFolder(exp_path)
                self.stop()
                return
            if len(values) > 1:
                select_folder.QtPrompt(
                    self, values, allow_abort=False, toggleMulti=True,
                    CbLabel="Select Position folder(s) to process:"
                )
                if select_folder.was_aborted:
                    self.logger.info(
                        'Process aborted by the user '
                        '(cancelled at Postion selection)'
                    )
                    self.stop()
                    return
                posFoldernames = select_folder.selected_pos
            else:
                posFoldernames = select_folder.pos_foldernames

        self.workerProgress(f'Selected folder: "{exp_path}"')
        self.workerProgress(' ')
        posListFormat = '\n'.join(posFoldernames)
        self.workerProgress(f'Selected Positions:\n{posListFormat}')
        self.workerProgress(' ')

        self.workerInitProgressBar(len(posFoldernames))
        
        self.thread = QThread()
        self.worker = workers.reapplyDataPrepWorker(exp_path, posFoldernames)

        self.worker.moveToThread(self.thread)
        self.worker.finished.connect(self.thread.quit)
        self.worker.finished.connect(self.worker.deleteLater)
        self.thread.finished.connect(self.thread.deleteLater)

        self.worker.progress.connect(self.workerProgress)
        self.worker.initPbar.connect(self.workerInitProgressBar)
        self.worker.updatePbar.connect(self.workerUpdateProgressBar)
        self.worker.critical.connect(self.workerCritical)
        self.worker.finished.connect(self.workerFinished)
        self.worker.sigCriticalNoChannels.connect(self.criticalNoChannelsFound)
        self.worker.sigSelectChannels.connect(self.selectChannels)

        self.thread.started.connect(self.worker.run)
        self.thread.start()
    
    def selectChannels(self, ch_name_selector, ch_names):
        win = widgets.QDialogListbox(
            'Select channel',
            'Select channel names to process:\n',
            ch_names, multiSelection=True, parent=self
        )
        win.exec_()
        if win.cancel:
            self.worker.abort = True
        self.worker.selectedChannels = win.selectedItemsText
        self.worker.waitCond.wakeAll()
    
    def criticalNotValidFolder(self, path: os.PathLike):
        txt = html_utils.paragraph(
            'The selected folder:<br><br>'
            f'<code>{path}</code><br><br>'
            'is <b>not a valid folder</b>. '
            'Select a folder that contains the Position_n folders'
        )
        msg = widgets.myMessageBox()
        msg.addShowInFileManagerButton(path)
        msg.critical(
            self, 'Incompatible folder', txt,
            buttonsTexts=('Ok',)
        )
    
    def criticalNoChannelsFound(self, images_path):
        err_title = 'Channel names not found'
        err_msg = html_utils.paragraph(
            'The following folder<br><br>'
            '<code>{images_path}</code><br><br>'
            '<b>does not valid channel files</b>.<br>'
        )
        msg = widgets.myMessageBox()
        msg.addShowInFileManagerButton(images_path)
        msg.critical(self, err_title, err_msg)
        self.logger.info(err_title)
        self.stop()
    
    def stop(self):
        self.startButton.show()
        self.stopButton.hide()

        if hasattr(self, 'worker'):
            self.worker.abort = True
    
    @exception_handler
    def workerInitProgressBar(self, maximum):
        self.progressBar.setValue(0)
        self.progressBar.setMaximum(maximum)
    
    @exception_handler
    def workerUpdateProgressBar(self):
        self.progressBar.update(1)
    
    @exception_handler
    def workerProgress(self, txt):
        self.logger.info(txt)
        self.logConsole.append(txt)
    
    @exception_handler
    def workerProgressBar(self, txt):
        self.logger.info(txt)
        self.logConsole.write(txt)
    
    @exception_handler
    def workerCritical(self, error):
        raise error
    
    @exception_handler
    def workerFinished(self):
        self.startButton.show()
        self.stopButton.hide()

        if self.worker.abort:
            msg = widgets.myMessageBox()
            msg.warning(
                self, 'Process stopped', 
                html_utils.paragraph('Data prep process stopped!')
            )
        else:
            msg = widgets.myMessageBox()
            msg.information(
                self, 'Process completed', 
                html_utils.paragraph('Data prep process completed!')
            )


from functools import wraps
import re
import sys
from matplotlib.colors import to_hex

is_mac = sys.platform == 'darwin'

def _tag(tag_info='p style="font-size:10px"'):
    def wrapper(func):
        @wraps(func)
        def inner(text):
            tag = tag_info.split(' ')[0]
            text = f'<{tag_info}>{text}</{tag}>'
            return text
        return inner
    return wrapper

def tag(text, tag_info='p style="font-size:10pt"'):
    tag = tag_info.split(' ')[0]
    text = f'<{tag_info}>{text}</{tag}>'
    return text

def href_tag(text, url):
    txt = tag(text, tag_info=f'a href="{url}"')
    return txt

def to_list(items, ordered=False):
    list_tag = 'ol' if ordered else 'ul'
    items_txt = ''.join([f'<li>{item}</li>' for item in items])
    txt = tag(items_txt, tag_info=list_tag)
    return txt

def span(text, color='r'):
    c = to_hex(color)
    return f'<span style="color: {c}">{text}</span>'

def css_head(txt):
    # if is_mac:
    #     txt = txt.replace(',', ',&nbsp;')
    s = (f"""
    <!DOCTYPE html>
    <html>
    <head>
    <style type="text/css">
            {txt}
    </style>
    </head>
    """)
    return s

def html_body(txt):
    if is_mac:
        txt = txt.replace(',', ',&nbsp;')
    s = (f"""
    <body>
        {txt}
    </body>
    </html>
    """)
    return s

def paragraph(txt, font_size='13px', font_color=None, wrap=True, center=False):
    # if is_mac:
    #     # Qt < 5.15.3 has a bug on macOS and the space after comma and perdiod
    #     # are super small. Force a non-breaking space (except for 'e.g.,').
    #     txt = txt.replace(',', ',&nbsp;')
    #     txt = txt.replace('.', '.&nbsp;')
    #     txt = txt.replace('e.&nbsp;g.&nbsp;', 'e.g.')
    #     txt = txt.replace('.&nbsp;.&nbsp;.&nbsp;', '...')
    #     txt = txt.replace('i.&nbsp;e.&nbsp;', 'i.e.')
    #     txt = txt.replace('etc.&nbsp;)', 'etc.)')
    if not wrap:
        txt = txt.replace(' ', '&nbsp;')
    if font_color is None:
        s = (f"""
        <p style="font-size:{font_size};">
            {txt}
        </p>
        """)
    else:
        s = (f"""
        <p style="font-size:{font_size}; color:{font_color}">
            {txt}
        </p>
        """)
    if center:
        s = re.sub(r'<p style="(.*)">', r'<p style="\1; text-align:center">', s)
    return s

# Syntax highlighting html
func_color = (111/255,66/255,205/255) # purplish
kwargs_color = (208/255,88/255,9/255) # reddish/orange
class_color = (215/255,58/255,73/255) # reddish
blue_color = (0/255,92/255,197/255) # blueish
class_sh = span('<i>class</i>', color=class_color)
def_sh = span('<i>def</i>', color=class_color)
if_sh = span('<i>if</i>', color=class_color)
elif_sh = span('<i>elif</i>', color=class_color)
kwargs_sh = span('**kwargs', color=kwargs_color)
Model_sh = span('Model', color=func_color)
segment_sh = span('segment', color=func_color)
predict_sh = span('predict', color=func_color)
CV_sh = span('CV', color=func_color)
init_sh = span('__init__', color=blue_color)
myModel_sh = span('MyModel', color=func_color)
return_sh = span('<i>return</i>', color=class_color)
equal_sh = span('=', color=class_color)
open_par_sh = span('(', color=blue_color)
close_par_sh = span(')', color=blue_color)
image_sh = span('image', color=kwargs_color)
from_sh = span('<i>from</i>', color=class_color)
import_sh = span('<i>import</i>', color=class_color)
is_not_sh = span('is not', color=class_color)
np_mean_sh = span('np.mean', color=class_color)
np_std_sh = span('np.std', color=class_color)


import os
from functools import partial
from natsort import natsorted

from .. import myutils, apps, load, printl, core, widgets
from ..utils import base

from PyQt5.QtCore import QTimer, pyqtSignal

class NapariArboretumDialog(base.MainThreadSinglePosUtilBase):
    def __init__(
            self, posPath, app, title: str, infoText: str, parent=None
        ):
        
        module = myutils.get_module_name(__file__)
        super().__init__(
            app, title, module, infoText, parent
        )

        self.sigClose.connect(self.close)

        func = partial(self.launchNapariArboretum, posPath)
        QTimer.singleShot(200, func)

    def launchNapariArboretum(self, posPath):
        images_path = os.path.join(posPath, 'Images')
        ls = myutils.listdir(images_path)

        image_files = [
            file for file in ls 
            if file.endswith('.tif') 
            or file.endswith('aligned.npz') 
            or file.endswith('.h5')
        ]

        selectImageFile = widgets.QDialogListbox(
            'Select image file',
            'Select which image file to load\n',
            image_files, multiSelection=False, parent=self
        )
        selectImageFile.exec_()
        if selectImageFile.cancel:
            self.logger.info('napari-arboretum utility aborted.')
            return

        imageFile = selectImageFile.selectedItemsText[0]
        self.logger.info(f'Loading image file {imageFile}...')
        
        imagePath = os.path.join(images_path, imageFile)
        posData = load.loadData(imagePath, '')
        posData.getBasenameAndChNames()
        posData.loadImgData()

        segm_files = load.get_segm_files(posData.images_path)
        existingEndnames = load.get_existing_segm_endnames(
            posData.basename, segm_files
        )

        if len(existingEndnames) > 1:
            win = apps.QDialogMultiSegmNpz(
                existingEndnames, images_path, parent=self, 
                basename=posData.basename
            )
            win.exec_()
            if win.cancel:
                self.logger.info('napari-arboretum utility aborted.')
                return
            selectedSegmEndName = win.selectedItemText
        else:
            selectedSegmEndName = existingEndnames[0]

        self.logger.info(f'Loading segmentation file ending with {selectedSegmEndName}...')

        posData.loadOtherFiles(
            load_segm_data=True,
            load_acdc_df=True,
            end_filename_segm=selectedSegmEndName
        )

        self.logger.info('Importing napari...')
        import napari

        self.logger.info('Building arboretum lineage tree...')
        acdc_df = posData.acdc_df.reset_index()
        tree = core.LineageTree(acdc_df)
        tracks_data, graph, properties = tree.to_arboretum()

        props = natsorted(acdc_df.columns.to_list())
        selectProps = widgets.QDialogListbox(
            'Select measurements',
            'Select measurements to add as <b>properties</b> in napari viewer<br><br>'
            '<code>Ctrl+Click</code> <i>to select multiple items</i><br>'
            '<code>Shift+Click</code> <i>to select a range of items</i><br>',
            props, multiSelection=True, parent=self
        )
        selectProps.exec_()
        if selectProps.cancel:
            self.logger.info('napari-arboretum utility aborted.')
            return
        
        for col in selectProps.selectedItemsText:
            try:
                properties[col] = acdc_df[col]
            except Exception as e:
                pass

        self.logger.info('Launching napari viewer...')
        viewer = napari.Viewer()
        viewer.add_image(posData.img_data, name=imageFile)
        viewer.add_labels(posData.segm_data, name=selectedSegmEndName)
        acdc_df_endname = selectedSegmEndName.replace('segm', 'acdc_tracks')
        viewer.add_tracks(
            tracks_data, graph=graph, name=acdc_df_endname, 
            properties=properties
        )
        viewer.window.add_plugin_dock_widget(
            plugin_name="napari-arboretum", widget_name="Arboretum"
        )

        napari.run(max_loop_level=2)

        self.logger.info('napari viewer closed.')
        self.close()

        



import numpy as np

def CV(
        signal, autoBkgr, dataPrepBkgr, objectRp, 
        correct_with_bkgr=False, which_bkgr='auto'
    ):
    """Function used to calculate coefficient of variation.

    NOTE: Make sure to call the function with the same name as the Python file
    containing this function (e.g., this file is called CV.py and the function
    is called CV)

    Parameters
    ----------
    signal : numpy 1D array
        This is a numpy array with all the intensities of the signal
        from each single segmented object.
    autoBkgr : single numeric value
        Median of all the background pixels (i.e. pixels with value 0 in the
        segmentation mask). Pass None if background correction with
        this value is not needed.
    dataPrepBkgr : single numeric value
        Median of all the pixels inside the background ROIs added during the
        data prep step (Cell-ACDC module 1).
        Pass None if background correction with this vaue is not needed.
    objectRp: skimage.measure.RegionProperties class
        Refer to `skimage.measure.regionprops` for more information
        on the available region properties.
    correct_with_bkgr : boolean
        Pass True if you need background correction.
    which_bkgr : string
        which_bkgr='auto' for correction with autoBkgr or
        which_bkgr='dataPrep' for correction with dataPrepBkgr

    Returns
    -------
    float
        Coefficient of Variation

    """
    if correct_with_bkgr:
        if which_bkgr=='auto':
            signal = signal - autoBkgr
        elif dataPrepBkgr is not None:
            signal = signal - dataPrepBkgr

    # Here goes your custom metric computation
    CV = np.std(signal)/np.mean(signal)

    return CV



import os
import cellacdc
from functools import partial

from PyQt5.QtWidgets import QDialog, QLabel, QGridLayout, QHBoxLayout
from PyQt5.QtGui import QPixmap
from PyQt5.QtCore import Qt

from ..myutils import read_version
from .. import widgets, myutils
from .. import qrc_resources

class QDialogAbout(QDialog):
    def __init__(self, parent=None):
        cellacdc_path = os.path.dirname(os.path.abspath(cellacdc.__file__))
        super().__init__(parent)
        self.setWindowFlags(Qt.Dialog)
        self.setWindowTitle('About Cell-ACDC')

        layout = QGridLayout()

        titleLabel = QLabel()
        txt = (f"""
        <p style="font-size:20px; font-family:ubuntu">
            <b>Cell-ACDC</b>
            <span style="font-size:12pt; font-family:ubuntu">
                (Analysis of the Cell Division Cycle)
            </span>
        </p>
        <p style="font-size:14px; font-family:ubuntu">
            Version {read_version()}
        </p>
        """)

        titleLabel.setText(txt)

        iconPixmap = QPixmap(":icon.ico")
        h = 128
        iconPixmap = iconPixmap.scaled(h,h, aspectRatioMode=Qt.KeepAspectRatio)
        iconLabel = QLabel()
        iconLabel.setPixmap(iconPixmap)

        github_url = r'https://github.com/SchmollerLab/Cell_ACDC'
        infoLabel = QLabel()
        infoLabel.setTextInteractionFlags(Qt.TextBrowserInteraction);
        infoLabel.setOpenExternalLinks(True);
        txt = (f"""
        <p style="font-size:13px; font-family:ubuntu">
            More info on our <a href=\"{github_url}">home page</a>.<br>
        </p>
        """)
        infoLabel.setText(txt)

        installedLayout = QHBoxLayout()
        installedLabel = QLabel()
        txt = (f"""
        <p style="font-size:12px; font-family:ubuntu">
            Installed in: {cellacdc_path}
        </p>
        """)
        installedLabel.setText(txt)

        button = widgets.showInFileManagerButton(
            myutils.get_open_filemaneger_os_string()
        )
        func = partial(myutils.showInExplorer, cellacdc_path)
        button.clicked.connect(func)
        installedLayout.addWidget(installedLabel)
        installedLayout.addStretch(1)
        installedLayout.addWidget(button)

        layout.addWidget(iconLabel, 0, 0)
        layout.addWidget(titleLabel, 0, 1, alignment=Qt.AlignLeft)
        layout.addWidget(infoLabel, 1, 1, alignment=Qt.AlignLeft)
        layout.setColumnStretch(2,1)
        layout.addLayout(installedLayout, 2, 0, 1, 3)

        self.setLayout(layout)

def _test():
    import sys
    from PyQt5.QtWidgets import QStyleFactory, QApplication
    app = QApplication(sys.argv)
    app.setStyle(QStyleFactory.create('Fusion'))
    win = QDialogAbout()
    win.show()
    app.exec_()



import numpy as np
import traceback
from skimage.measure import label, regionprops, regionprops_table
import skimage.transform
import pandas as pd
from tifffile import imread
import os
import glob
from math import pow, floor
from tqdm import tqdm
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QApplication, QStyleFactory, QFileDialog
from PyQt5 import QtCore
import sys
import difflib
from scipy.stats import binned_statistic
import warnings

from . import myutils, prompts, apps, qrc_resources, widgets, html_utils, printl

def configuration_dialog():
    if os.name == 'nt':
        try:
            # Set taskbar icon in windows
            import ctypes
            myappid = 'schmollerlab.cellacdc.pyqt.v1' # arbitrary string
            ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)
        except Exception as e:
            pass
    # app = QtCore.QCoreApplication.instance()
    # if app is None:
    app = QApplication(sys.argv)
    app.setStyle(QStyleFactory.create('Fusion'))
    app.setWindowIcon(QIcon(":icon.ico"))
    continue_selection = True
    data_dirs = []
    positions = []
    while continue_selection:
        MostRecentPath = myutils.getMostRecentPath()
        data_dir = QFileDialog.getExistingDirectory(
            None, 'Select experiment folder containing Position_n folders ',
            MostRecentPath
        )
        if not data_dir:
            continue_selection = False
            break

        myutils.addToRecentPaths(data_dir)
        foldername = os.path.basename(data_dir)
        if foldername == 'Images':
            pos_path = os.path.dirname(data_dir)
            data_dir = os.path.dirname(pos_path)
            pos = [os.path.basename(pos_path)]
        elif foldername.find('Position_') != -1:
            pos_path = data_dir
            data_dir = os.path.dirname(data_dir)
            pos = [os.path.basename(pos_path)]
        else:
            available_pos = myutils.get_pos_foldernames(data_dir)
            if not available_pos:
                print('******************************')
                print('Selected folder does not contain any Position folders.')
                print(f'Selected folder: "{data_dir}"')
                print('******************************')
                raise FileNotFoundError

            win = widgets.QDialogListbox(
                'Position Selection',
                'Select which position(s) you want to analyse',
                available_pos
            )
            win.show()
            win.exec_()
            if win.cancel:
                print('******************************')
                print('Execution aborted by the user')
                print('******************************')
                raise InterruptedError
            pos = win.selectedItemsText

        data_dirs.append(data_dir)
        positions.append(pos)
        msg = widgets.myMessageBox()
        txt = html_utils.paragraph(
            'Do you wish to select Positions from other experiments?'
        )
        yes, no = msg.question(
            None, 'Continue selection?', txt, buttonsTexts=(' Yes ', ' No ')
        )
        continue_selection = msg.clickedButton == yes
    if len(data_dirs) == 0:
        print('******************************')
        print("No positions selected!")
        print('******************************')
        raise IndexError
    return data_dirs, positions

def find_available_channels(filenames, first_pos_dir):
    ch_name_selector = prompts.select_channel_name()
    ch_names, warn = ch_name_selector.get_available_channels(
        filenames, first_pos_dir
    )
    return ch_names, warn

def calculate_downstream_data(
    file_names,
    image_folders,
    positions,
    channels,
    force_recalculation=False
):
    no_of_channels = len(channels)
    overall_df = pd.DataFrame()
    for file_idx, file in enumerate(file_names):
        for pos_idx, pos_dir in enumerate(image_folders[file_idx]):
            channel_data = ('placeholder')*no_of_channels
            print(f'Load files for {file}, {positions[file_idx][pos_idx]}...')
            try:
                *channel_data, seg_mask, cc_data, metadata, cc_props = _load_files(pos_dir, channels)
            except TypeError:
                print(f'File {file}, position {positions[file_idx][pos_idx]} skipped due to missing segmentation mask/CC annotations.')
                continue
            print(f'Number of cells in position: {len(cc_data.Cell_ID.unique())}')
            print(f'Number of annotated frames in position: {cc_data.frame_i.max()+1}')
            cc_data = _rename_columns(cc_data)
            is_timelapse_data, is_zstack_data = False, False
            if int(metadata.loc['SizeT'])>1:
                is_timelapse_data=True
            if int(metadata.loc['SizeZ'])>1:
                is_zstack_data=True
            if cc_props is not None and not force_recalculation:
                print('Cell Cycle property data already existing, loaded from disk...')
                overall_df = pd.concat([overall_df,cc_props], ignore_index=True).reset_index(drop=True)
            else:
                print(f'Calculate regionprops on each frame based on Segmentation...')
                rp_df = _calculate_rp_df(seg_mask, is_timelapse_data, is_zstack_data, metadata, max_frame=cc_data.frame_i.max()+1)
                print(f'Calculate signal metrics for every channel and cell...')
                flu_signal_df = _calculate_flu_signal(
                    seg_mask,
                    channel_data,
                    channels,
                    cc_data,
                    is_timelapse_data,
                    is_zstack_data
                )
                temp_df = cc_data.merge(
                    rp_df, on=['frame_i', 'Cell_ID'], how='left',
                    suffixes=('_gui', '')
                )
                temp_df = temp_df.merge(
                    flu_signal_df, on=['frame_i', 'Cell_ID'], how='left',
                    suffixes=('_gui', '')
                )
                # calculate amount of corrected signal by multiplying mean with area
                if is_timelapse_data:
                    for channel in channels:
                        temp_df[f'{channel}_corrected_amount'] = (
                            temp_df[f'{channel}_corrected_mean']
                            * temp_df['area']
                        )
                        try:
                            temp_df[f'{channel}_corrected_concentration'] = (
                                temp_df[f'{channel}_corrected_amount']
                                / temp_df['cell_vol_fl']
                            )
                        except KeyError:
                            print(f'Volume is missing in acdc output, NaNs inserted in concentration columns of channel {channel}')
                            temp_df[f'{channel}_corrected_concentration'] = None
                temp_df['max_frame_pos'] = cc_data.frame_i.max()
                temp_df['file'] = file
                temp_df['selection_subset'] = file_idx
                temp_df['position'] = positions[file_idx][pos_idx]
                temp_df['directory'] = pos_dir
                print('Saving calculated data for next time...')
                files_in_curr_dir = myutils.listdir(pos_dir)
                common_prefix = _determine_common_prefix(files_in_curr_dir)
                save_path = os.path.join(pos_dir, f'{common_prefix}cca_properties_downstream.csv')
                temp_df.to_csv(save_path, index=False)
                overall_df = pd.concat([overall_df, temp_df], ignore_index=True).reset_index(drop=True)

    print('Done!')
    return overall_df, is_timelapse_data, is_zstack_data


def calculate_relatives_data(overall_df, channels):
    # Join on Cell_ID vs. relative_ID to later calculate columns like "daughter growth" or "mother-bud-signal-combined"
    overall_df_rel = overall_df.copy()
    overall_df = overall_df.merge(
        overall_df_rel,
        how='left',
        left_on=['frame_i', 'relative_ID', 'max_frame_pos', 'file', 'selection_subset', 'position', 'directory'],
        right_on=['frame_i', 'Cell_ID', 'max_frame_pos', 'file', 'selection_subset', 'position', 'directory'],
        suffixes = ('', '_rel')
    )
    # for every channel, calculate amount from mother and bud cells combined
    for ch in channels:
        try:
            overall_df[f'{ch}_combined_amount_mother_bud'] = overall_df.apply(
                lambda x: (
                    x.loc[f'{ch}_corrected_amount']
                    + x.loc[f'{ch}_corrected_amount_rel']
                    if x.loc['cell_cycle_stage']=='S'
                    else x.loc[f'{ch}_corrected_amount']
                    ),
                axis=1
            )
            overall_df[f'{ch}_combined_raw_sum_mother_bud'] = overall_df.apply(
                lambda x: (
                    x.loc[f'{ch}_raw_sum']
                    + x.loc[f'{ch}_raw_sum_rel']
                    if x.loc['cell_cycle_stage']=='S'
                    else x.loc[f'{ch}_raw_sum']
                    ),
                axis=1
            )
        except KeyError:
            continue
    overall_df['combined_mother_bud_volume'] = overall_df.apply(
        lambda x: x.loc['cell_vol_fl']+x.loc['cell_vol_fl_rel'] if\
        x.loc['cell_cycle_stage']=='S' else\
        x.loc['cell_vol_fl'],
        axis=1
    )
    return overall_df


def calculate_per_phase_quantities(overall_df, group_cols, channels):
    # group by group columns, aggregate some other columns
    phase_grouped = overall_df.sort_values(
        'frame_i'
    ).groupby(group_cols).agg(
        # perform some calculations relating to the whole phase:
        phase_area_growth=('cell_area_um2', lambda x: x.iloc[-1]-x.iloc[0]),
        phase_volume_growth=('cell_vol_fl', lambda x: x.iloc[-1]-x.iloc[0]),
        phase_area_at_beginning=('cell_area_um2', 'first'),
        phase_volume_at_beginning=('cell_vol_fl', 'first'),
        phase_volume_at_end=('cell_vol_fl', 'last'),
        phase_daughter_area_growth=('cell_area_um2_rel', lambda x: x.iloc[-1]-x.iloc[0]),
        phase_daughter_volume_growth=('cell_vol_fl_rel', lambda x: x.iloc[-1]-x.iloc[0]),
        phase_length=('frame_i', lambda x: max(x)-min(x)),
        phase_begin = ('frame_i', 'min'),
        phase_end = ('frame_i', 'max'),
        phase_combined_volume_at_end = ('combined_mother_bud_volume','last')
    ).reset_index()
    # calculate some quantities in a for loop for all available channels and merge results.
    phase_grouped_flu = pd.DataFrame(columns=group_cols)
    for ch in channels:
        if f'{ch}_corrected_mean' in overall_df.columns:
            flu_temp = overall_df.sort_values(
                'frame_i'
            ).groupby(group_cols).agg({
                # perform some calculations on flu data:
                f'{ch}_corrected_amount': 'first',
                f'{ch}_corrected_mean': 'first',
                f'{ch}_corrected_concentration': ['first','last'],
                f'{ch}_combined_amount_mother_bud': ['first','last']
            }).reset_index()
            # collapse multiindex into column name with aggregation as suffix
            flu_temp.columns = ['_'.join(col) if col[1]!='' else col[0] for col in flu_temp.columns.values]
            # rename columns into meaningful names
            flu_temp = flu_temp.rename({
                f'{ch}_corrected_amount_first': f'phase_{ch}_amount_at_beginning',
                f'{ch}_corrected_mean_first': f'phase_{ch}_mean_at_beginning',
                f'{ch}_corrected_concentration_first': f'phase_{ch}_concentration_at_beginning',
                f'{ch}_corrected_concentration_last': f'phase_{ch}_concentration_at_end',
                f'{ch}_combined_amount_mother_bud_first': f'phase_{ch}_combined_amount_at_beginning',
                f'{ch}_combined_amount_mother_bud_last': f'phase_{ch}_combined_amount_at_end',
            }, axis=1)
            phase_grouped_flu = phase_grouped_flu.merge(flu_temp, how='right', on=group_cols, suffixes=('',''))

    # detect complete cell cycle phases and complete cell cycles
    temp = np.logical_and(
        phase_grouped.phase_begin > 0,
        phase_grouped.phase_end < phase_grouped.max_frame_pos
    )
    # this or is for disappearing cells
    if 'max_t' in overall_df.columns:
        complete_phase_indices = np.logical_and(
            temp,
            phase_grouped.phase_end < phase_grouped.max_t
        )
    else:
        complete_phase_indices = temp
    phase_grouped['complete_phase'] = complete_phase_indices.astype(int)
    no_of_compl_phases_per_cycle = phase_grouped.groupby(
        ['Cell_ID', 'generation_num', 'position', 'file']
    )['complete_phase'].transform('sum')
    complete_cycle_indices = no_of_compl_phases_per_cycle == 2
    phase_grouped['complete_cycle'] = complete_cycle_indices.astype(int)
    # join phase-grouped data with
    phase_grouped = phase_grouped.merge(phase_grouped_flu, how='left', on=group_cols, suffixes=('',''))
    return phase_grouped


def _determine_common_prefix(filenames):
    basename = filenames[0]
    for file in filenames:
        # Determine the basename based on intersection of all .tif
        _, ext = os.path.splitext(file)
        sm = difflib.SequenceMatcher(None, file, basename)
        i, j, k = sm.find_longest_match(0, len(file),
                                        0, len(basename))
        basename = file[i:i+k]
    return basename


def _auto_rescale_intensity(img, perc=0.01, clip_min=False):
    """
    function to clip outliers to the given percentiles and scale afterwards.

    use skimage.exposure.rescale_intensity for clipping all images in the
    given percentiles to the percentile border.

    Parameters
    ----------
    img: np.array
        image to eliminate outliers from
    perc: float
        percentage of pixels which are considered as outliers.
        Example: given 0.01, the lowest and highest 1% of pixels are considered
        as outliers and clipped to the 1- and 99-percentile respectively.

    Returns
    -------
    scaled_img: np.array
        image, where outlier pixels are set to 1- and 99-percentiles and which is
        scaled to [0,1] afterwards
    """
    if perc > 0:
        vmin, vmax = np.percentile(img, q=(perc, 100-perc))
        clip_min_indices = img < vmin
        clip_max_indices = img > vmax
        if clip_min:
            img[clip_min_indices] = vmin
        img[clip_max_indices] = vmax
    else:
        vmin, vmax = img.min(), img.max()
    scaled_img = (img-vmin)/(vmax-vmin)
    return scaled_img


def _load_files(file_dir, channels):
    """
    Function to load files of all given channels and the corresponding segmentation masks.
    Check first if aligned files are available and use them if so.
    """
    no_of_aligned_files = len(glob.glob(os.path.join(f'{file_dir}', '*aligned.npz')))
    seg_mask_available = len(glob.glob(os.path.join(f'{file_dir}', '*_segm.npz'))) > 0
    acdc_output_available = (
        len(glob.glob(os.path.join(f'{file_dir}', '*acdc_output.csv')))
        + len(glob.glob(os.path.join(f'{file_dir}', '*cc_stage*'))) > 0
    )
    if not (seg_mask_available and acdc_output_available):
        return None
    channel_files = []
    if no_of_aligned_files > 0:
        for channel in channels:
            try:
                ch_aligned_path = glob.glob(os.path.join(f'{file_dir}', f'*{channel}_aligned.npz'))[0]
                channel_files.append(np.load(ch_aligned_path)['arr_0'])
            except IndexError:
                try:
                    ch_aligned_path = glob.glob(os.path.join(f'{file_dir}', f'*{channel}_aligned.npy'))[0]
                    channel_files.append(np.load(ch_aligned_path))
                except IndexError:
                    print(f'Could not find an aligned file for channel {channel}')
                    print(f'Resulting data will not contain fluorescent data for this channel')
                    channel_files.append(None)
    else:
        for channel in channels:
            try:
                ch_not_aligned_path = glob.glob(os.path.join(f'{file_dir}', f'*{channel}*'))[0]
                channel_files.append(imread(ch_not_aligned_path))
            except IndexError:
                print(f'Could not find any file for channel {channel}')
                print(f'Resulting data will not contain fluorescent data for this channel')
                channel_files.append(None)

    # append segmentation file
    try:
        segm_file_path = glob.glob(os.path.join(f'{file_dir}', '*_segm.npz'))[0]
        channel_files.append(np.load(segm_file_path)['arr_0'])
    except IndexError:
        segm_file_path = glob.glob(os.path.join(f'{file_dir}', '*_segm.npy'))[0]
        # assume segmentation mask to be .npy
        channel_files.append(np.load(segm_file_path))
    # append cc-data
    try:
        cc_stage_path = glob.glob(os.path.join(f'{file_dir}', '*acdc_output.csv'))[0]
    except IndexError:
        cc_stage_path = glob.glob(os.path.join(f'{file_dir}', '*cc_stage.csv'))[0]
    # assume cell cycle output of ACDC to be .csv
    channel_files.append(pd.read_csv(cc_stage_path))

    # append metadata if available, else append None
    if len(glob.glob(os.path.join(f'{file_dir}', '*metadata*'))) > 0:
        metadata_path = glob.glob(os.path.join(f'{file_dir}', '*metadata.csv'))[0]
        # assume calculated metadata to be .csv
        channel_files.append(pd.read_csv(metadata_path).set_index('Description'))
    else:
        channel_files.append(None)

    # append cc-properties if available, else append None
    if len(glob.glob(os.path.join(f'{file_dir}', '*_downstream*'))) > 0:
        cc_props_path = glob.glob(os.path.join(f'{file_dir}', '*_downstream*'))[0]
        # assume calculated cc properties to be .csv
        channel_files.append(pd.read_csv(cc_props_path))
    else:
        channel_files.append(None)
    return tuple(channel_files)




def _calculate_rp_df(seg_mask, is_timelapse_data, is_zstack_data, metadata, max_frame=1, label_input=False):
    """
    function to calculate regionprops based on a 2D(!) segmentation mask.
    TODO: insert check if 3D segmentation mask is available and calculate more regionprops.
    """
    if label_input:
        #generate labeled video only when input is not labeled yet
        labeled_data = label(seg_mask)
    else:
        labeled_data = seg_mask.copy()
    # calculate rp's for rings
    t_df = pd.DataFrame()
    props = ('label', 'area', 'convex_area', 'filled_area','major_axis_length',
             'minor_axis_length', 'orientation', 'perimeter', 'centroid', 'solidity')
    rename_dict = {'label':'Cell_ID', 'centroid-0':'centroid_y', 'centroid-1':'centroid_x'}
    if is_timelapse_data:
        for t, img in enumerate(tqdm(labeled_data)):
            # build time-dependent dataframes for further use (later for cca)
            if img.max() > 0:
                t_rp_df = pd.DataFrame(regionprops_table(img.astype(int), properties=props)).rename(columns=rename_dict)
                t_rp_df['frame_i'] = t
                # calculate volumes based on regionprops
                if metadata is None:
                    warnings.warn("No metadata available. Volumes are not calculated")
                    t_rp_df['cell_vol_vox_downstream'] = 0
                    t_rp_df['cell_vol_fl_downstream'] = 0
                else:
                    t_rp = regionprops(img.astype(int))
                    vol_vox = [_calc_rot_vol(obj, metadata.loc["PhysicalSizeY"], metadata.loc["PhysicalSizeX"])[0] for obj in t_rp]
                    vol_fl = [_calc_rot_vol(obj, metadata.loc["PhysicalSizeY"], metadata.loc["PhysicalSizeX"])[1] for obj in t_rp]
                    assert len(t_rp_df) == len(vol_vox)
                    t_rp_df['cell_vol_vox_downstream'] = vol_vox
                    t_rp_df['cell_vol_fl_downstream'] = vol_fl
                # determine id's which are falsely merged by 3D-labeling
                for r_id in t_rp_df.Cell_ID.unique():
                    bin_label = label((img==r_id).astype(int))
                    t_rp_df.loc[t_rp_df['Cell_ID']==r_id, '2d_label_count'] = bin_label.max()
                t_df = pd.concat([t_df, t_rp_df], ignore_index=True)
        # calculate global features by grouping
        grouped_df = t_df.groupby('Cell_ID').agg(
            min_t=('frame_i', min),
            max_t=('frame_i', max),
            lifespan=('frame_i', lambda x: max(x)-min(x)+1)
        ).reset_index()
        merged_df = t_df.merge(grouped_df, how='left', on='Cell_ID')
        # calculate further indicators based on merged data
        merged_df['age'] = merged_df['frame_i'] - merged_df['min_t'] + 1
        merged_df['frames_till_gone'] = merged_df['max_t'] - merged_df['frame_i']
        merged_df['elongation'] = merged_df['major_axis_length']/merged_df['minor_axis_length']
        return merged_df
    else:
        rp_df = pd.DataFrame(regionprops_table(labeled_data.astype(int), properties=props)).rename(columns=rename_dict)
        for r_id in rp_df.Cell_ID.unique():
            bin_label = label((labeled_data==r_id).astype(int))
            rp_df.loc[rp_df['Cell_ID']==r_id, '2d_label_count'] = bin_label.max()
        rp_df['elongation'] = rp_df['major_axis_length']/rp_df['minor_axis_length']
        rp_df['frame_i'] = 0
        return rp_df


def _calc_rot_vol(obj, PhysicalSizeY=1, PhysicalSizeX=1, logger=None):
    """Given the region properties of a 2D object (from skimage.measure.regionprops).
    calculate the rotation volume as described in the Supplementary information of
    https://www.nature.com/articles/s41467-020-16764-x

    Parameters
    ----------
    obj : class skimage.measure.RegionProperties
        Single item of the list returned by from skimage.measure.regionprops.
    PhysicalSizeY : type
        Physical size of the pixel in the Y-diretion in micrometer/pixel.
    PhysicalSizeX : type
        Physical size of the pixel in the X-diretion in micrometer/pixel.

    Returns
    -------
    tuple
        Tuple of the calculate volume in voxels and femtoliters.

    Notes
    -------
    For 3D objects we take the z-slice passing through the centroid

    We convert PhysicalSizeY and PhysicalSizeX to float because when they are
    read from csv they might be a string value.

    """
    is3Dobj = False
    try:
        orientation = obj.orientation
    except NotImplementedError as e:
        # if obj.image.ndim != 3:
        #     printl(e, obj.image.ndim, obj.bbox, obj.centroid)
        is3Dobj = True

    try:
        if is3Dobj:
            # For 3D objects we use a max projection for the rotation
            obj_lab = obj.image.max(axis=0).astype(np.uint32)*obj.label
            obj = regionprops(obj_lab)[0]

        vox_to_fl = float(PhysicalSizeY)*pow(float(PhysicalSizeX), 2)
        rotate_ID_img = skimage.transform.rotate(
            obj.image.astype(np.single), -(obj.orientation*180/np.pi),
            resize=True, order=3
        )
        radii = np.sum(rotate_ID_img, axis=1)/2
        vol_vox = np.sum(np.pi*(radii**2))
        if vox_to_fl is not None:
            return vol_vox, float(vol_vox*vox_to_fl)
        else:
            return vol_vox, vol_vox
    except Exception as e:
        if logger is not None:
            logger.exception(traceback.format_exc())
        else:
            printl(traceback.format_exc())
        return np.nan, np.nan


def _calculate_flu_signal(seg_mask, channel_data, channels, cc_data, is_timelapse_data, is_zstack_data):
    """
    function to calculate sum and scaled sum of fluorescence signal per frame and cell.
    channel_data is a list-like of TYX arrays, one for each channel.
    channels are the name of the channels in the tuple.
    cc_data the output of acdc.
    """
    max_frame = cc_data.frame_i.max()
    df = pd.DataFrame(columns=['frame_i', 'Cell_ID'])
    bg_medians = []
    for ch_idx, ch_array in enumerate(channel_data):
        if ch_array is None:
            bg_medians.append(None)
        else:
            bg_index = np.logical_and(seg_mask[:max_frame+1]==0, ch_array[:max_frame+1]!=0)
            ch_medians = [np.median(ch_array[t][bg_index[t]]) for t in range(max_frame+1)]
            bg_medians.append(ch_medians)
    if is_timelapse_data:
        for cell_id in tqdm(cc_data.Cell_ID.unique()):
            temp_df = pd.DataFrame(columns=['frame_i', 'Cell_ID'])
            times = range(max_frame+1)
            temp_df['frame_i'] = times; temp_df['Cell_ID'] = cell_id
            index_array = (seg_mask[:max_frame+1] == cell_id)
            channel_data_cut = [c_arr[:max_frame+1] if c_arr is not None else None for c_arr in channel_data]
            for c_idx, c_array in enumerate(channel_data_cut):
                if c_array is not None:
                    cell_signal = c_array*index_array
                    # cell_signal = c_array[index_array]
                    summed = np.sum(cell_signal, axis=(1,2))
                    # count = np.sum(cell_signal!=0, axis=(1,2))
                    count = np.sum(index_array, axis=(1,2))
                    mean_signal = np.divide(summed, count, where=count!=0)
                    # mean_signal = np.mean(cell_signal, axis=(1,2))
                    corrected_signal = mean_signal - np.array(bg_medians[c_idx])
                    temp_df[f'{channels[c_idx]}_corrected_mean'] = corrected_signal
                    temp_df[f'{channels[c_idx]}_raw_sum'] = summed
                else:
                    temp_df[f'{channels[c_idx]}_corrected_mean'] = 0
                    temp_df[f'{channels[c_idx]}_raw_sum'] = 0
            df = pd.concat([df, temp_df], ignore_index=True)
        signal_indices = np.array(['_corrected_mean' in col for col in df.columns])
        keep_rows = df.loc[:,signal_indices].sum(axis=1) > 0
        df = df[keep_rows]
        df = df.sort_values(['frame_i', 'Cell_ID']).reset_index(drop=True)
    return df


def _rename_columns(cc_data):
    rename_dict = {
        'Cell cycle stage': 'cell_cycle_stage',
        '# of cycles': 'generation_num',
        "Relative's ID": 'relative_ID',
        'Relationship': 'relationship',
        'Emerg_frame_i': 'emerg_frame_i',
        'Division_frame_i': 'division_frame_i',
        'Discard': 'is_cell_excluded'
    }
    cc_data.columns = [rename_dict.get(col, col) for col in cc_data.columns]
    return cc_data


def binned_mean_stats(x, values, nbins, bins_min_count):
    """
    function to calculate binned means and corresponding standard errors for
    evenly spaced bins in the data ("x" gets distributed in bins, stats are calculated on "values")
    """
    bin_counts, _, _ = binned_statistic(x, values, statistic='count', bins=nbins)
    bin_means, bin_edges, _ = binned_statistic(x, values, bins=nbins)
    bin_std, _, _ = binned_statistic(x, values, statistic='std', bins=nbins)
    bin_standard_errors = bin_std/np.sqrt(bin_counts)
    bin_width = (bin_edges[1] - bin_edges[0])
    bin_centers = bin_edges[1:] - bin_width/2
    x_errorbar = bin_centers[bin_counts>bins_min_count]
    y_errorbar = bin_means[bin_counts>bins_min_count]
    err_errorbar = 1.96 * bin_standard_errors[bin_counts>bins_min_count]
    return x_errorbar, y_errorbar, err_errorbar


def calculate_effect_size_cohen(data, group1, group2, cat_column='size_category', val_column='Pp38_concentration'):
    assert cat_column in data.columns and val_column in data.columns
    data_gr1 = data[data[cat_column]==group1]
    data_gr2 = data[data[cat_column]==group2]
    n1 = len(data_gr1)
    n2 = len(data_gr2)
    s1 = np.var(data_gr1[val_column])
    s2 = np.var(data_gr2[val_column])
    cohen_s = np.sqrt(
        ((n1-1)*s1+(n2-1)*s2) / (n1+n2-2)
    )
    effect_size = (np.mean(data_gr1[val_column])- np.mean(data_gr2[val_column])) / cohen_s
    return effect_size

def calculate_effect_size_glass(data, group1, group2, cat_column='size_category', val_column='Pp38_concentration'):
    assert cat_column in data.columns and val_column in data.columns
    data_gr1 = data[data[cat_column]==group1]
    data_gr2 = data[data[cat_column]==group2]
    glass_s = np.std(data_gr2[val_column])
    effect_size = (np.mean(data_gr1[val_column])- np.mean(data_gr2[val_column])) / glass_s
    return effect_size



import sys
import os
import shutil
import re
import traceback
import time
import datetime
import tempfile
import h5py
import difflib
import pathlib
import numpy as np
import pandas as pd
from collections import Counter
from tqdm import tqdm
from natsort import natsorted
from pprint import pprint
from functools import wraps, partial
from itertools import permutations

from PyQt5.QtWidgets import (
    QApplication, QMainWindow, QFileDialog,
    QVBoxLayout, QPushButton, QLabel, QStyleFactory,
    QWidget, QMessageBox, QPlainTextEdit, QProgressBar
)
from PyQt5.QtCore import (
    Qt, QObject, pyqtSignal, QThread, QMutex, QWaitCondition,
    QEventLoop
)
from PyQt5 import QtGui

# Here we use from cellacdc because this script is laucnhed in
# a separate process that doesn't have a parent package
from . import exception_handler
from . import qrc_resources
from . import apps, myutils, widgets, html_utils, printl

if os.name == 'nt':
    try:
        # Set taskbar icon in windows
        import ctypes
        myappid = 'schmollerlab.cellacdc.pyqt.v1' # arbitrary string
        ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)
    except:
        pass

def worker_exception_handler(func):
    @wraps(func)
    def run(self):
        try:
            result = func(self)
        except Exception as error:
            result = None
            self.critical.emit(error)
        return result
    return run

class bioFormatsWorker(QObject):
    finished = pyqtSignal()
    progress = pyqtSignal(str)
    progressPbar = pyqtSignal(int)
    initPbar = pyqtSignal(int)
    criticalError = pyqtSignal(str, str, str)
    filesExisting = pyqtSignal(str)
    confirmMetadata = pyqtSignal(
        str, float, str, int, int, int, int,
        float, str, float, float, float,
        str, list, list, str, str, object
    )
    critical = pyqtSignal(object)
    # aborted = pyqtSignal()

    def __init__(
            self, raw_src_path, rawFilenames, exp_dst_path,
            mutex, waitCond, rawDataStruct
        ):
        QObject.__init__(self)
        self.raw_src_path = raw_src_path
        self.exp_dst_path = exp_dst_path
        self.rawFilenames = rawFilenames
        self.mutex = mutex
        self.waitCond = waitCond
        self.askReplacePosFiles = True
        self.overWriteMetadata = False
        self.trustMetadataReader = False
        self.rawDataStruct = rawDataStruct
        self.overwritePos = False
        self.addFiles = False
        self.cancel = False

    def readMetadata(self, raw_src_path, filename):
        rawFilePath = os.path.join(raw_src_path, filename)

        self.progress.emit('Reading OME metadata...')

        try:
            metadataXML = bioformats.get_omexml_metadata(rawFilePath)
            metadata = bioformats.OMEXML(metadataXML)
            self.metadata = metadata
            self.metadataXML = metadataXML
        except Exception as e:
            self.isCriticalError = True
            self.criticalError.emit(
                'reading image data or metadata',
                traceback.format_exc(), filename
            )
            return True

        try:
            LensNA = float(metadata.instrument().Objective.LensNA)
        except Exception as e:
            self.progress.emit(
                '===================================================')
            self.progress.emit(rawFilePath)
            self.progress.emit('WARNING: LensNA not found in metadata.')
            self.progress.emit(
                '===================================================')
            LensNA = 1.4

        if self.rawDataStruct != 2:
            try:
                SizeS = int(metadata.get_image_count())
            except Exception as e:
                self.progress.emit(
                    '===================================================')
                self.progress.emit(rawFilePath)
                self.progress.emit('WARNING: SizeS not found in metadata.')
                self.progress.emit(
                    '===================================================')
                SizeS = 1
        else:
            SizeS = self.SizeS

        DimensionOrder = 'zct'

        try:
            SizeZ = int(metadata.image().Pixels.SizeZ)
        except Exception as e:
            self.progress.emit(
                '===================================================')
            self.progress.emit(rawFilePath)
            self.progress.emit('WARNING: SizeZ not found in metadata.')
            self.progress.emit(
                '===================================================')
            SizeZ = 1

        try:
            SizeT = int(metadata.image().Pixels.SizeT)
        except Exception as e:
            self.progress.emit(
                '===================================================')
            self.progress.emit(rawFilePath)
            self.progress.emit('WARNING: SizeT not found in metadata.')
            self.progress.emit(
                '===================================================')
            SizeT = 1

        try:
            Pixels = metadata.image().Pixels
            TimeIncrement = float(Pixels.node.get('TimeIncrement'))
        except Exception as e:
            self.progress.emit(
                '===================================================')
            self.progress.emit(rawFilePath)
            self.progress.emit('WARNING: TimeIncrement not found in metadata.')
            self.progress.emit(
                '===================================================')
            TimeIncrement = 1.0

        try:
            Pixels = metadata.image().Pixels
            TimeIncrementUnit = Pixels.node.get('TimeIncrementUnit')
            if TimeIncrementUnit is None:
                raise
        except Exception as e:
            self.progress.emit(
                '===================================================')
            self.progress.emit(rawFilePath)
            self.progress.emit('WARNING: TimeIncrementUnit not found in metadata.')
            self.progress.emit(
                '===================================================')
            TimeIncrementUnit = 's'

        try:
            SizeC = int(metadata.image().Pixels.SizeC)
        except Exception as e:
            self.progress.emit(
                '===================================================')
            self.progress.emit(rawFilePath)
            self.progress.emit('WARNING: SizeC not found in metadata.')
            self.progress.emit(
                '===================================================')
            SizeC = 1

        try:
            PhysicalSizeX = float(metadata.image().Pixels.PhysicalSizeX)
        except Exception as e:
            self.progress.emit(
                '===================================================')
            self.progress.emit(rawFilePath)
            self.progress.emit('WARNING: PhysicalSizeX not found in metadata.')
            self.progress.emit(
                '===================================================')
            PhysicalSizeX = 1.0

        try:
            PhysicalSizeY = float(metadata.image().Pixels.PhysicalSizeY)
        except Exception as e:
            self.progress.emit(
                '===================================================')
            self.progress.emit(rawFilePath)
            self.progress.emit('WARNING: PhysicalSizeY not found in metadata.')
            self.progress.emit(
                '===================================================')
            PhysicalSizeY = 1.0

        try:
            PhysicalSizeZ = float(metadata.image().Pixels.PhysicalSizeZ)
        except Exception as e:
            self.progress.emit(
                '===================================================')
            self.progress.emit(rawFilePath)
            self.progress.emit('WARNING: PhysicalSizeZ not found in metadata.')
            self.progress.emit(
                '===================================================')
            PhysicalSizeZ = 1.0

        try:
            Pixels = metadata.image().Pixels
            PhysicalSizeUnit = Pixels.node.get('PhysicalSizeXUnit')
            if PhysicalSizeUnit is None:
                raise
        except Exception as e:
            self.progress.emit(
                '===================================================')
            self.progress.emit(rawFilePath)
            self.progress.emit('WARNING: PhysicalSizeUnit not found in metadata.')
            self.progress.emit(
                '===================================================')
            PhysicalSizeUnit = 'μm'

        try:
            ImageName = metadata.image().Name
            if ImageName is None:
                raise
        except Exception as e:
            self.progress.emit(
                '===================================================')
            self.progress.emit(rawFilePath)
            self.progress.emit('WARNING: Image Name not found in metadata.')
            self.progress.emit(
                '===================================================')
            ImageName = ''


        if self.rawDataStruct != 2:
            try:
                chNames = ['name_not_found']*SizeC
                for c in range(SizeC):
                    try:
                        chNames[c] = metadata.image().Pixels.Channel(c).Name
                    except Exception as e:
                        pass
            except Exception as e:
                self.progress.emit(
                    '===================================================')
                self.progress.emit(rawFilePath)
                self.progress.emit('WARNING: chNames not found in metadata.')
                self.progress.emit(
                    '===================================================')
                chNames = ['not_found']*SizeC
        else:
            chNames = self.chNames
            SizeC = len(self.chNames)

        if self.rawDataStruct != 2:
            try:
                emWavelens = [500.0]*SizeC
                for c in range(SizeC):
                    Channel = metadata.image().Pixels.Channel(c)
                    emWavelen = Channel.node.get("EmissionWavelength")
                    try:
                        emWavelens[c] = float(emWavelen)
                    except Exception as e:
                        emWavelens[c] = 0.0
            except Exception as e:
                traceback.print_exc()
                self.progress.emit(
                    '===================================================')
                self.progress.emit(rawFilePath)
                self.progress.emit('WARNING: EmissionWavelength not found in metadata.')
                self.progress.emit(
                    '===================================================')
                emWavelens = [500.0]*SizeC
        else:
            emWavelens = [500.0]*SizeC

        if self.trustMetadataReader:
            self.LensNA = LensNA
            self.DimensionOrder = DimensionOrder
            self.SizeT = SizeT
            self.SizeZ = SizeZ
            self.SizeC = SizeC
            self.SizeS = SizeS
            self.TimeIncrement = TimeIncrement
            self.PhysicalSizeX = PhysicalSizeX
            self.PhysicalSizeY = PhysicalSizeY
            self.PhysicalSizeZ = PhysicalSizeZ
            # self.chNames = chNames
            self.emWavelens = emWavelens
        else:
            self.mutex.lock()
            sampleImgData = None
            if self.rawDataStruct != 2:
                sampleImgData = {}
                self.progress.emit('Reading sample image data...')
                dimsIdx = {}
                if SizeT >= 4:
                    sampleSizeT = 4
                else:
                    sampleSizeT = SizeT  
                with bioformats.ImageReader(rawFilePath) as reader:
                    for dimsOrd in permutations('zct', 3):
                        allChannelsData = []
                        idxs = self.buildIndexes(SizeC, SizeT, SizeZ, dimsOrd)
                        for c in range(SizeC):
                            dimsIdx['c'] = c
                            imgData_tz = []
                            for t in range(sampleSizeT):   
                                dimsIdx['t'] = t
                                imgData_z = []
                                for z in range(SizeZ):
                                    dimsIdx['z'] = z
                                    idx = self.getIndex(idxs, dimsIdx, dimsOrd)
                                    imgData = reader.read(
                                        c=c, z=z, t=0, rescale=False, index=idx
                                    )
                                    imgData_z.append(imgData)                
                                imgData_z = np.array(imgData_z, dtype=imgData.dtype)
                                imgData_z = np.squeeze(imgData_z)
                                imgData_tz.append(imgData_z)
                            imgData_tz = np.array(imgData_tz, dtype=imgData.dtype)
                            imgData_tz = np.squeeze(imgData_tz)
                            allChannelsData.append(imgData_tz)
                        sampleImgData[''.join(dimsOrd)] = allChannelsData
            
            self.confirmMetadata.emit(
                filename, LensNA, DimensionOrder, SizeT, SizeZ, SizeC, SizeS,
                TimeIncrement, TimeIncrementUnit, PhysicalSizeX, PhysicalSizeY,
                PhysicalSizeZ, PhysicalSizeUnit, chNames, emWavelens, ImageName,
                rawFilePath, sampleImgData
            )
            self.waitCond.wait(self.mutex)
            self.mutex.unlock()
            if self.metadataWin.cancel:
                return True
            elif self.metadataWin.overWrite:
                self.overWriteMetadata = True
            elif self.metadataWin.trust:
                self.trustMetadataReader = True

            self.to_h5 = self.metadataWin.to_h5
            self.LensNA = self.metadataWin.LensNA
            self.DimensionOrder = self.metadataWin.DimensionOrder
            self.SizeT = self.metadataWin.SizeT
            self.SizeZ = self.metadataWin.SizeZ
            self.SizeC = self.metadataWin.SizeC
            self.SizeS = self.metadataWin.SizeS
            self.TimeIncrement = self.metadataWin.TimeIncrement
            self.PhysicalSizeX = self.metadataWin.PhysicalSizeX
            self.PhysicalSizeY = self.metadataWin.PhysicalSizeY
            self.PhysicalSizeZ = self.metadataWin.PhysicalSizeZ
            self.selectedPos = self.metadataWin.selectedPos
            self.chNames = self.metadataWin.chNames
            self.saveChannels = self.metadataWin.saveChannels
            self.emWavelens = self.metadataWin.emWavelens
            self.addImageName = self.metadataWin.addImageName

    def saveToPosFolder(
            self, p, raw_src_path, exp_dst_path, filename, series, p_idx=0
        ):
        rawFilePath = os.path.join(raw_src_path, filename)

        if os.path.basename(raw_src_path) == 'raw_microscopy_files':
            raw_src_path = os.path.dirname(raw_src_path)
        
        pos_name = f'Position_{p+1}'
        savePos = (
            'All Positions' in self.selectedPos or pos_name in self.selectedPos
        )
        if not savePos:
            return False

        pos_path = os.path.join(exp_dst_path, f'Position_{p+1}')
        images_path = os.path.join(pos_path, 'Images')

        if os.path.exists(images_path) and self.askReplacePosFiles:
            self.askReplacePosFiles = False
            self.mutex.lock()
            self.filesExisting.emit(pos_path)
            self.waitCond.wait(self.mutex)
            self.mutex.unlock()
            if self.cancel:
                return True

        if os.path.exists(images_path) and self.overwritePos:
            shutil.rmtree(images_path)
        
        if not os.path.exists(images_path):
            os.makedirs(images_path)

        self.saveData(images_path, rawFilePath, filename, p, series, p_idx=p_idx)

        return False

    def removeInvalidCharacters(self, chName_in):
        # Remove invalid charachters
        chName = "".join(
            c if c.isalnum() or c=='_' or c=='' else '_' for c in chName_in
        )
        trim_ = chName.endswith('_')
        while trim_:
            chName = chName[:-1]
            trim_ = chName.endswith('_')

    def getFilename(
            self, filenameNOext, s0p, appendTxt, series, ext, 
            return_basename=False
        ):
        # Do not allow dots in the filename since it breaks stuff here and there
        filenameNOext = filenameNOext.replace('.', '_')
        if self.addImageName:
            try:
                ImageName = self.metadata.image(index=series).Name
                if not isinstance(ImageName, str):
                    raise
            except Exception as e:
                ImageName = ''
            self.removeInvalidCharacters(ImageName)
            basename = f'{filenameNOext}_{ImageName}_s{s0p}_'
            filename = f'{basename}{appendTxt}{ext}'
        else:
            basename = f'{filenameNOext}_s{s0p}_'
            filename = f'{basename}{appendTxt}{ext}'
        if return_basename:
            return filename, basename
        else:
            return filename
    
    def buildIndexes(self, SizeC, SizeT, SizeZ, DimensionOrder):
        SizesCTZ = {'c': SizeC, 't': SizeT, 'z': SizeZ}
        idxs = {}
        k_key, i_key, j_key = DimensionOrder
        idx = 0

        for k in range(SizesCTZ[k_key]):
            for i in range(SizesCTZ[i_key]):
                for j in range(SizesCTZ[j_key]):
                    idxs[(k,i,j)] = idx
                    idx += 1
        return idxs

    def getIndex(self, idxs, dimsIdx, DimensionOrder):
        """Get the index of the single 2D image given `dimsIdx`. 
        
        Note that `idxs` is generated in `buildIndexes` method.

        Example:
            Given a `DimensionOrder = 'tcz'`, and a 
            `dimsIdx = {'t': 0, 'c': 1, 'z': 0}` the returned index is `3`.

        Args:
            idxs (dict): Dictionary where the keys are tuples of `(t, c, z)`, 
                i.e., `t` is the index requested in the time dimension.
                Values are the corresponding incremental indexes in the correct
                order (see `DimensionOrder`)
            dimsIdx (dict): Dictionary with three keys, 'z', 't', and 'c'. 
                Values are the corresponding requested index.
            DimensionOrder (str): String of three lower-case characters 
                (no spaces, no punctuation) from a combination of 'z', 't', 
                and 'c'. This value determines the order of dimensions in the 
                Bio-Formats file. 

        Returns:
            int: incremental index requested.
        

        """
        dims = tuple([dimsIdx.get(v, 0) for v in DimensionOrder])
        return idxs[dims]
            
    def saveImgDataChannel(
            self, reader, series, images_path, filenameNOext, s0p, chName,
            ch_idx, idxs
        ):
        if self.to_h5:
            filename = self.getFilename(
                filenameNOext, s0p, chName, series, '.h5'
            )
            tempDir = tempfile.mkdtemp()
            tempFilepath = os.path.join(tempDir, filename)
            print('==========================================================')
            print(f'.h5 tempfile: "{tempFilepath}"')
            print('==========================================================')
            h5f = h5py.File(tempFilepath, 'w')
            # Read SizeX and SizeY from the shape of one image
            imgData = reader.read(
                c=ch_idx, z=0, t=0, series=series, rescale=False
            )
            shape = (self.SizeT, self.SizeZ, *imgData.shape)
            chunks = (1,1,*imgData.shape)
            imgData_ch = h5f.create_dataset(
                'data', shape, dtype=imgData.dtype,
                chunks=chunks, shuffle=False
            )
        else:
            filename = self.getFilename(
                filenameNOext, s0p, chName, series, '.tif'
            )
            imgData_ch = []

        filePath = os.path.join(images_path, filename)
        dimsIdx = {'c': ch_idx} 
        for t in range(self.SizeT):
            imgData_z = []
            dimsIdx['t'] = t
            for z in range(self.SizeZ):
                dimsIdx['z'] = z
                if self.rawDataStruct != 2:
                    idx = self.getIndex(idxs, dimsIdx, self.DimensionOrder)
                else:
                    idx = None
                imgData = reader.read(
                    c=ch_idx, z=z, t=t, series=series, rescale=False,
                    index=idx
                )
                if self.to_h5:
                    imgData_ch[t, z] = imgData
                else:
                    imgData_z.append(imgData)

            if not self.to_h5:
                imgData_z = np.squeeze(np.array(imgData_z, dtype=imgData.dtype))
                imgData_ch.append(imgData_z)

        if not self.to_h5:
            imgData_ch = np.squeeze(np.array(imgData_ch, dtype=imgData.dtype))
            myutils.imagej_tiffwriter(
                filePath, imgData_ch, {}, self.SizeT, self.SizeZ
            )
        else:
            h5f.close()
            shutil.move(tempFilepath, filePath)
            shutil.rmtree(tempDir)

    def saveData(self, images_path, rawFilePath, filename, p, series, p_idx=0):
        s0p = str(p+1).zfill(self.numPosDigits)
        self.progress.emit(
            f'Position {p+1}/{self.numPos}: saving data to {images_path}...'
        )
        filenameNOext, ext = os.path.splitext(filename)

        metadataXML_path = os.path.join(
            images_path,
            self.getFilename(filenameNOext, s0p, 'metadataXML', series, '.txt')
        )
        with open(metadataXML_path, 'w', encoding="utf-8") as txt:
            txt.write(self.metadataXML)

        metadata_filename, basename = self.getFilename(
            filenameNOext, s0p, 'metadata', series, '.csv', 
            return_basename=True
        )
        metadata_csv_path = os.path.join(images_path, metadata_filename)
        df = pd.DataFrame({
            'LensNA': self.LensNA,
            'DimensionOrder': self.DimensionOrder,
            'SizeT': self.SizeT,
            'SizeZ': self.SizeZ,
            'TimeIncrement': self.TimeIncrement,
            'PhysicalSizeZ': self.PhysicalSizeZ,
            'PhysicalSizeY': self.PhysicalSizeY,
            'PhysicalSizeX': self.PhysicalSizeX,
            'basename': basename
        }, index=['values']).T
        df.index.name = 'Description'

        ch_metadata = [
            chName for c, chName in enumerate(self.chNames)
            if self.saveChannels[c]
        ]
        description = [
            f'channel_{c}_name' for c in range(self.SizeC) 
            if self.saveChannels[c]
        ]
        ch_metadata.extend([
            wavelen for c, wavelen in enumerate(self.emWavelens)
            if self.saveChannels[c]
        ])
        description.extend([
            f'channel_{c}_emWavelen' for c in range(self.SizeC)
            if self.saveChannels[c]
        ])

        df_channelNames = pd.DataFrame({
            'Description': description,
            'values': ch_metadata
        }).set_index('Description')

        df = pd.concat([df, df_channelNames])

        if os.path.exists(metadata_csv_path):
            # Keep channel names already existing and not saved now
            existing_df = pd.read_csv(metadata_csv_path).set_index('Description')
            for c, chName in enumerate(self.chNames):
                if self.saveChannels[c]:
                    continue
                chName_idx = f'channel_{c}_name'
                chWavelen_idx = f'channel_{c}_emWavelen'
                try:
                    existing_chName = existing_df.at[chName_idx, 'values']
                    df.at[chName_idx, 'values'] = existing_chName
                except Exception as e:
                    traceback.print_exc()
                    pass
                
                try:
                    existing_chWavelen = existing_df.at[chWavelen_idx, 'values']
                    df.at[chWavelen_idx, 'values'] = existing_chWavelen
                except Exception as e:
                    traceback.print_exc()
                    pass

        df.to_csv(metadata_csv_path)

        idxs = self.buildIndexes(
            self.SizeC, self.SizeT, self.SizeZ, self.DimensionOrder
        )
        if self.rawDataStruct != 2:       
            with bioformats.ImageReader(rawFilePath) as reader:
                iter = enumerate(zip(self.chNames, self.saveChannels))
                for c, (chName, saveCh) in iter:
                    self.progressPbar.emit(1)
                    if not saveCh:
                        continue

                    self.progress.emit(
                        f'  Saving channel {c+1}/{len(self.chNames)} ({chName})'
                    )
                    self.saveImgDataChannel(
                        reader, series, images_path, filenameNOext, s0p,
                        chName, c, idxs
                    )

        elif self.rawDataStruct == 2:
            iter = enumerate(zip(self.chNames, self.saveChannels))
            pos_rawFilenames = []
            basename = filename
            for c, (chName, saveCh) in iter:
                self.progressPbar.emit(1)
                if not saveCh:
                    continue

                rawFilename = f'{basename}{p+1}_{chName}'
                pos_rawFilenames.append(rawFilename)
                raw_src_path = os.path.dirname(rawFilePath)
                rawFilePath = [
                    os.path.join(raw_src_path, f)
                    for f in myutils.listdir(raw_src_path)
                    if f.find(rawFilename)!=-1
                ][0]

                with bioformats.ImageReader(rawFilePath) as reader:
                    self.progress.emit(
                        f'  Saving channel {c+1}/{len(self.chNames)} ({chName})'
                    )
                    imgData_ch = []
                    self.saveImgDataChannel(
                        reader, series, images_path, filenameNOext, s0p,
                        chName, 0, idxs
                    )

            if self.moveOtherFiles or self.copyOtherFiles:
                # Move the other files present in the folder if they
                # contain "otherFilename" in the name
                otherFilename = f'{basename}{p+1}'
                rawFilePath = set()
                for f in myutils.listdir(raw_src_path):
                    notRawFile = all(
                        [f.find(rawName)==-1 for rawName in pos_rawFilenames]
                    )
                    isPosFile = f.find(otherFilename)!=-1
                    if isPosFile and notRawFile:
                        rawFilePath.add(os.path.join(raw_src_path, f))

                for file in rawFilePath:
                    # Determine basename, posNum and chName to build
                    # filename as "basename_s01_chName.ext"
                    _filename = os.path.basename(file)
                    m = re.findall(fr'{basename}(\d+)_(.+)', _filename)
                    if not m or len(m[0])!=2:
                        dst = os.path.join(images_path, _filename)
                    else:
                        _chNameWithExt = m[0][1]
                        _filename = f'{filenameNOext}_s{s0p}_{_chNameWithExt}'
                        dst = os.path.join(images_path, _filename)
                    if self.moveOtherFiles:
                        try:
                            shutil.move(file, dst)
                        except Exception as e:
                            self.progress.emit(e)
                    elif self.copyOtherFiles:
                        try:
                            shutil.copy(file, dst)
                        except Exception as e:
                            self.progress.emit(e)

    @worker_exception_handler
    def run(self):
        raw_src_path = self.raw_src_path
        exp_dst_path = self.exp_dst_path
        javabridge.start_vm(class_path=bioformats.JARS, run_headless=True)
        # bioformats.init_logger()
        self.progress.emit('Java VM running.')
        self.aborted = False
        self.isCriticalError = False
        for p, filename in enumerate(self.rawFilenames):
            if self.rawDataStruct == 0:
                if not self.overWriteMetadata:
                    abort = self.readMetadata(raw_src_path, filename)
                    if abort:
                        self.aborted = True
                        break

                self.numPos = self.SizeS
                self.numPosDigits = len(str(self.numPos))
                if p == 0:
                    self.initPbar.emit(self.numPos*self.SizeC)
                for p in range(self.SizeS):
                    abort = self.saveToPosFolder(
                        p, raw_src_path, exp_dst_path, filename, p
                    )
                    if abort:
                        self.aborted = True
                        break

            elif self.rawDataStruct == 1:
                if not self.overWriteMetadata:
                    abort = self.readMetadata(raw_src_path, filename)
                    if abort:
                        self.aborted = True
                        break
                self.numPos = len(self.rawFilenames)
                self.numPosDigits = len(str(self.numPos))
                if p == 0:
                    self.initPbar.emit(self.numPos*self.SizeC)
                abort = self.saveToPosFolder(
                    p, raw_src_path, exp_dst_path, filename, 0
                )
                if abort:
                    self.aborted = True
                    break

            else:
                break

            # Move files to raw_microscopy_files folder
            foldername = os.path.basename(self.raw_src_path)
            if foldername != 'raw_microscopy_files' and not self.aborted:
                rawFilePath = os.path.join(self.raw_src_path, filename)
                raw_path = os.path.join(raw_src_path, 'raw_microscopy_files')
                if not os.path.exists(raw_path):
                    os.mkdir(raw_path)
                dst = os.path.join(raw_path, filename)
                try:
                    shutil.move(rawFilePath, dst)
                except PermissionError as e:
                    self.progress.emit(e)

        if self.rawDataStruct == 2:
            filename = self.rawFilenames[0]
            if not self.overWriteMetadata:
                abort = self.readMetadata(raw_src_path, filename)
                if abort:
                    self.aborted = True
                    javabridge.kill_vm()
                    self.finished.emit()
                    return


            self.numPos = len(self.posNums)
            self.numPosDigits = len(str(self.numPos))
            self.initPbar.emit(self.numPos*self.SizeC)
            for p_idx, pos in enumerate(self.posNums):
                p = pos-1
                abort = self.saveToPosFolder(
                    p, raw_src_path, exp_dst_path, self.basename,
                    0, p_idx=p_idx
                )
                if abort:
                    self.aborted = True
                    break

            for filename in self.rawFilenames:
                # Move files to raw_microscopy_files folder
                foldername = os.path.basename(self.raw_src_path)
                if foldername != 'raw_microscopy_files' and not self.aborted:
                    rawFilePath = os.path.join(self.raw_src_path, filename)
                    raw_path = os.path.join(raw_src_path, 'raw_microscopy_files')
                    if not os.path.exists(raw_path):
                        os.mkdir(raw_path)
                    dst = os.path.join(raw_path, filename)
                    try:
                        shutil.move(rawFilePath, dst)
                    except PermissionError as e:
                        self.progress.emit(e)

        javabridge.kill_vm()
        self.finished.emit()
        

class createDataStructWin(QMainWindow):
    sigFinishedReadingSampleImgData = pyqtSignal(object)

    def __init__(
            self, parent=None, allowExit=False,
            buttonToRestore=None, mainWin=None,
            start_JVM=True, version=None
        ):
        super().__init__(parent)

        self._version = version

        logger, logs_path, log_path, log_filename = myutils.setupLogger(
            module='dataStruct'
        )
        self.logger = logger
        self.log_path = log_path
        self.log_filename = log_filename
        self.logs_path = logs_path

        if self._version is not None:
            logger.info(f'Initializing Data structure module v{self._version}...')
        else:
            logger.info(f'Initializing Data structure module...')

        
        is_linux = sys.platform.startswith('linux')
        is_mac = sys.platform == 'darwin'
        is_win = sys.platform.startswith("win")

        self.start_JVM = start_JVM
        self.allowExit = allowExit
        self.processFinished = False
        self.buttonToRestore = buttonToRestore
        self.mainWin = mainWin

        self.setWindowTitle("Cell-ACDC - From raw microscopy file to tifs")
        self.setWindowIcon(QtGui.QIcon(":icon.ico"))

        mainContainer = QWidget()
        self.setCentralWidget(mainContainer)

        mainLayout = QVBoxLayout()

        label = QLabel(
            'Creating data structure from raw microscopy file(s)...'
        )

        label.setStyleSheet("padding:5px 10px 10px 10px;")
        label.setAlignment(Qt.AlignCenter)
        font = QtGui.QFont()
        font.setPixelSize(14)
        font.setBold(True)
        label.setFont(font)
        mainLayout.addWidget(label)

        informativeHtml = (
        """
        <html>
        <head>
        <title></title>
        <style type="text/css">
        blockquote {
         margin: 5;
         padding: 0;
        }
        </style>
        </head>
        <body>
        <blockquote>
        <p style="font-size:14px; line-height:1.2">
            This <b>wizard</b> will guide you through the <b>creation of the required
            data structure</b><br> starting from the raw microscopy file(s)
        </p>
        <p style="font-size:12px; line-height:1.2">
            Follow the instructions in the pop-up windows.<br>
            Note that pop-ups might be minimized or behind other open windows.<br>
            Keep an eye on the terminal/console in case of any error.
        </p>
        <p style="font-size:12px; line-height:1.2">
            Progress will be displayed below.
        </p>
        </blockquote>
        </body>
        </html>
        """
        )

        informativeText = QLabel(self)

        informativeText.setTextFormat(Qt.RichText)
        informativeText.setText(informativeHtml)
        informativeText.setStyleSheet("padding:5px 0px 10px 0px;")
        mainLayout.addWidget(informativeText)

        self.logWin = QPlainTextEdit()
        self.logWin.setReadOnly(True)
        mainLayout.addWidget(self.logWin)

        abortButton = widgets.cancelPushButton(' Abort process ')
        abortButton.clicked.connect(self.close)
        mainLayout.addWidget(abortButton, alignment=Qt.AlignCenter)

        mainLayout.setContentsMargins(20, 0, 20, 20)
        mainContainer.setLayout(mainLayout)

        self.mainLayout = mainLayout

        if not is_win and not is_mac:
            if parent is None:
                self.show()
            self.criticalOSnotSupported()
            self.close()
            raise OSError('This module is supported ONLY on Windows 10/10 and macOS')

        global bioformats, javabridge
        self.logger.info('Checking if Java is installed...')
        try:
            import javabridge
        except ModuleNotFoundError as e:
            print('======================================')
            traceback_str = traceback.format_exc()
            self.logger.exception(traceback_str)
            print('======================================')
            cancel = myutils.install_javabridge_help(parent=self)
            if cancel:
                raise ModuleNotFoundError(
                    'User aborted javabridge installation'
                )

            isGitInstalled = myutils.check_git_installed(parent=self)
            if not isGitInstalled:
                raise ModuleNotFoundError(
                    'Git is not installed. Install from '
                    'https://git-scm.com/book/en/v2/Getting-Started-Installing-Git'
                )

            try:
                jre_path, jdk_path, url = myutils.download_java()
            except Exception as e:
                print('======================================')
                traceback_str = traceback.format_exc()
                self.logger.exception(traceback_str)
                print('======================================')
                java_info = myutils.get_java_url()
                url, file_size, os_foldername, unzipped_foldername = java_info
                acdc_java_path, _ = myutils.get_acdc_java_path()
                java_href = f'<a href="{url}">this</a>'
                s = (
                    f'1. Download {java_href} .zip file and unzip it.<br>'
                    '2. Inside the unzipped folder there should be a folder called '
                    f'"{unzipped_foldername}". Open that folder and copy its '
                    'content to the following path:<br><br>'
                    f'{os.path.join(acdc_java_path, os_foldername)}'
                )
                note = (
                    '<br><br><i>NOTE: if clicking on the link above does not work '
                    'copy the link below and paste it into the browser</i><br><br>'
                    f'{url}'
                )
                msg = QMessageBox()
                msg.setIcon(msg.Warning)
                msg.setWindowTitle('Java not found')
                msg.setTextFormat(Qt.RichText)

                txt = (f"""
                <p style=font-size:13px>
                    This module requires Java to work.<br><br>
                    Follow the instructions below and then try
                    launching this module again.<br><br>
                    {s}{note}
                </p>
                """)
                msg.setText(txt)
                msg.exec_()
                err = s.replace('<br>', ' ')
                err = err.replace('<a href=', '')
                err = err.replace('>this</a>', '')
                raise ModuleNotFoundError(
                    'Installation of module "javabridge" failed. '
                    f'{err}'
                )

            if not is_win:
                cancel = myutils.install_java()
                if cancel:
                    raise ModuleNotFoundError(
                        'User aborted Java installation'
                    )
                    return

            myutils.install_javabridge()

        except Exception as e:
            print('======================================')
            traceback_str = traceback.format_exc()
            self.logger.exception(traceback_str)
            print('======================================')
            cancel = myutils.install_java()
            if cancel:
                raise ModuleNotFoundError(
                    'User aborted Java installation'
                )
                return
            myutils.install_javabridge(
                force_compile=True, attempt_uninstall_first=True
            )

        try:
            import javabridge
            from cellacdc import bioformats
        except Exception as e:
            print('===============================================================')
            traceback_str = traceback.format_exc()
            self.logger.exception(traceback_str)
            error_msg = (
                'Error while importing "javabridge" and "bioformats"\n\n'
                'Please report detailed error (click "Show details") '
                'here: https://github.com/SchmollerLab/Cell_ACDC/issues'
            )
            print(error_msg)
            print('===============================================================')

            msg = QMessageBox(self)
            msg.setWindowTitle('import javabridge/bioformats error')
            msg.setIcon(msg.Critical)
            msg.setText(error_msg)
            msg.setDetailedText(traceback.format_exc())
            msg.exec_()
            raise FileNotFoundError('Dowload of Java failed. See above for details.')

    def criticalOSnotSupported(self):
        from cellacdc import widgets
        if self.parent() is None:
            msg = widgets.myMessageBox(self)
        else:
            msg = widgets.myMessageBox(self.parent())
        msg.setIcon(iconName='SP_MessageBoxCritical')
        msg.setWindowTitle('Not a supported OS')
        msg.addButton('    Ok     ')
        err_msg = (f"""
        <p style="font-size:12px">
        Unfortunately, the module "0. Create data structure from microscopy file(s)"
        is functional <b>only on Windows 10/11 and macOS</b>.<br><br>
        We are working on extending support to other Operating Systems.<br><br>
        Please open an issue on our
        <a href="https://github.com/SchmollerLab/Cell_ACDC/issues">
            GitHub page
        </a> to request this feature.<br>
        Note that <b>all other modules are functional</b> on
        macOS, Linux and Windows.<br><br>
        In the meantine, to create the required data structure,
        you can use an <b>automated Fiji macro</b> that you can download from
        <a href="https://github.com/SchmollerLab/Cell_ACDC/tree/main/FijiMacros">
            here
        </a>.<br><br>
        Check out the <b>instructions</b> on how to use the macros
        in the section  <b>"Create data structure using Fiji Macros"</b> of the
        user manual.<br><br>
        You can download the user manual from
        <a href="https://github.com/SchmollerLab/Cell_ACDC/blob/main/UserManual/Cell-ACDC_User_Manual.pdf">
            here
        </a>.
        </p>
        """)
        msg.addText(err_msg)
        # msg_label = msg.findChild(QLabel, "qt_msgbox_label")
        # msg_label.setOpenExternalLinks(False)
        # msg_label.linkActivated.connect(self.on_linkActivated)
        msg.exec_()


    def on_linkActivated(self, link):
        if link == 'manual':
            systems = {
                'nt': os.startfile,
                'posix': lambda foldername: os.system('xdg-open "%s"' % foldername),
                'os2': lambda foldername: os.system('open "%s"' % foldername)
                 }

            main_path = pathlib.Path(__file__).resolve().parents[1]
            userManual_path = main_path / 'UserManual'
            systems.get(os.name, os.startfile)(userManual_path)
        elif link == 'fiji':
            systems = {
                'nt': os.startfile,
                'posix': lambda foldername: os.system('xdg-open "%s"' % foldername),
                'os2': lambda foldername: os.system('open "%s"' % foldername)
                 }

            main_path = pathlib.Path(__file__).resolve().parents[1]
            fijiMacros_path = main_path / 'FijiMacros'
            systems.get(os.name, os.startfile)(fijiMacros_path)


    def getMostRecentPath(self):
        cellacdc_path = os.path.dirname(os.path.realpath(__file__))
        recentPaths_path = os.path.join(
            cellacdc_path, 'temp', 'recentPaths.csv'
        )
        if os.path.exists(recentPaths_path):
            df = pd.read_csv(recentPaths_path, index_col='index')
            if 'opened_last_on' in df.columns:
                df = df.sort_values('opened_last_on', ascending=False)
            self.MostRecentPath = df.iloc[0]['path']
            if not isinstance(self.MostRecentPath, str):
                self.MostRecentPath = ''
        else:
            self.MostRecentPath = ''

    def addToRecentPaths(self, raw_src_path):
        if not os.path.exists(raw_src_path):
            return
        cellacdc_path = os.path.dirname(os.path.realpath(__file__))
        recentPaths_path = os.path.join(
            cellacdc_path, 'temp', 'recentPaths.csv'
        )
        if os.path.exists(recentPaths_path):
            df = pd.read_csv(recentPaths_path, index_col='index')
            recentPaths = df['path'].to_list()
            if 'opened_last_on' in df.columns:
                openedOn = df['opened_last_on'].to_list()
            else:
                openedOn = [np.nan]*len(recentPaths)
            if raw_src_path in recentPaths:
                pop_idx = recentPaths.index(raw_src_path)
                recentPaths.pop(pop_idx)
                openedOn.pop(pop_idx)
            recentPaths.insert(0, raw_src_path)
            openedOn.insert(0, datetime.datetime.now())
            # Keep max 20 recent paths
            if len(recentPaths) > 20:
                recentPaths.pop(-1)
                openedOn.pop(-1)
        else:
            recentPaths = [raw_src_path]
            openedOn = [datetime.datetime.now()]
        df = pd.DataFrame({'path': recentPaths,
                           'opened_last_on': pd.Series(openedOn,
                                                       dtype='datetime64[ns]')})
        df.index.name = 'index'
        df.to_csv(recentPaths_path)

    @exception_handler
    def main(self):
        self.log('Asking how raw data is structured...')
        rawDataStruct, abort = self.askRawDataStruct()
        if abort:
            self.close()
            return

        self.rawDataStruct = rawDataStruct

        if rawDataStruct == 3:
            self.instructManualStruct()
            self.close()
            return

        self.log('Instructing to move raw data...')
        proceed = self.instructMoveRawFiles()
        if not proceed:
            self.close()
            return

        self.log(
            'Asking to select the folder that contains the microscopy files...'
        )
        self.getMostRecentPath()
        raw_src_path = QFileDialog.getExistingDirectory(
            self, 'Select folder containing the microscopy files', self.MostRecentPath)
        self.addToRecentPaths(raw_src_path)

        if raw_src_path == '':
            self.close()
            return

        self.log(
            'Checking file format of loaded files...'
        )
        rawFilenames = self.checkFileFormat(raw_src_path)
        if not rawFilenames:
            self.close()
            return

        if rawDataStruct == 2:
            proceed = self.attemptSeparateMultiChannel(rawFilenames)
            if not proceed:
                self.close()
                return

        self.log(
            'Asking in which folder to save the images files...'
        )
        exp_dst_path = QFileDialog.getExistingDirectory(
            self, 'Select the folder in which to save the images files',
            raw_src_path
        )
        if not exp_dst_path:
            self.close()
            return

        self.addToRecentPaths(exp_dst_path)

        self.log(
            'Starting a Java Virtual Machine...'
        )

        self.addPbar()

        # Set up separate thread for bioFormatsWorker class
        self.mutex = QMutex()
        self.waitCond = QWaitCondition()
        self.thread = QThread()
        self.worker = bioFormatsWorker(
            raw_src_path, rawFilenames, exp_dst_path,
            self.mutex, self.waitCond, rawDataStruct
        )
        if self.rawDataStruct == 2:
            self.worker.basename = self.basename
            self.worker.SizeS = self.SizeS
            self.worker.posNums = self.posNums
            self.worker.chNames = self.chNames
            self.worker.moveOtherFiles = self.moveOtherFiles
            self.worker.copyOtherFiles = self.copyOtherFiles

        self.worker.moveToThread(self.thread)

        self.worker.finished.connect(self.thread.quit)
        self.worker.finished.connect(self.worker.deleteLater)
        self.worker.finished.connect(self.taskEnded)
        self.thread.finished.connect(self.thread.deleteLater)
        self.worker.initPbar.connect(self.setPbarMax)
        self.worker.progressPbar.connect(self.updatePbar)
        self.worker.progress.connect(self.log)
        self.worker.critical.connect(self.workerCritical)
        self.worker.criticalError.connect(self.criticalBioFormats)
        self.worker.confirmMetadata.connect(self.askConfirmMetadata)
        self.worker.filesExisting.connect(self.askReplacePos)
        self.thread.started.connect(self.worker.run)

        self.thread.start()

    @exception_handler
    def workerCritical(self, error):
        raise error

    def instructManualStruct(self):
        issues_url = 'https://github.com/SchmollerLab/Cell_ACDC/issues'
        manual_url = 'https://github.com/SchmollerLab/Cell_ACDC/blob/main/UserManual/Cell-ACDC_User_Manual.pdf'
        txt = (
        f"""
        <p style="font-size:11px;">
        If you would like to add compatibility with your raw microscopy files,<br>
        you can request a new feature <a href=\"{issues_url}">here</a>.<br><br>
        Please label the issue as "enhancement" and provide as many details as
        possible about how your raw microscopy files are structured.<br><br>
        A second option is to create the required data structure manually.
        Please have a look at the instruction on the
        <a href=\"{manual_url}">User manual</a> at the section called
        "Manually create data structure from microscopy file(s)"
        </p>
        """
        )
        msg = QMessageBox(self)
        msg.setWindowTitle('Data structure not available')
        msg.setIcon(msg.Information)
        msg.setText(txt)
        msg.setTextInteractionFlags(Qt.TextBrowserInteraction)
        msg.setTextFormat(Qt.RichText)
        msg.exec_()

    def addPbar(self):
        self.QPbar = QProgressBar(self)
        self.QPbar.setValue(0)
        palette = QtGui.QPalette()
        palette.setColor(QtGui.QPalette.Highlight, QtGui.QColor(207, 235, 155))
        palette.setColor(QtGui.QPalette.Text, QtGui.QColor(0, 0, 0))
        palette.setColor(QtGui.QPalette.HighlightedText, QtGui.QColor(0, 0, 0))
        self.QPbar.setPalette(palette)
        self.mainLayout.insertWidget(3, self.QPbar)

    def updatePbar(self, deltaPbar):
        self.QPbar.setValue(self.QPbar.value()+deltaPbar)

    def setPbarMax(self, max):
        self.QPbar.setMaximum(max)

    def taskEnded(self):
        if self.worker.aborted and not self.worker.isCriticalError:
            msg = QMessageBox(self)
            abort = msg.critical(
               self, 'Conversion task aborted.',
               'Conversion task aborted.',
               msg.Close
            )
            self.close()
        elif not self.worker.aborted:
            msg = QMessageBox(self)
            abort = msg.information(
               self, 'Conversion task ended.',
               'Conversion task ended.\n\n'
               f'Files saved to "{self.worker.exp_dst_path}"',
               msg.Close
            )
            self.close()

    def log(self, text):
        self.logWin.appendPlainText(text)

    def askRawDataStruct(self):
        infoText =  html_utils.paragraph(
            'Select how you have your <b>raw microscopy files arranged</b>'
        )
        win = apps.QDialogCombobox(
            'Raw data structure',
            [
                'Single microscopy file with one or more positions',
                'Multiple microscopy files, one for each position',
                'Multiple microscopy files, one for each channel',
                'NONE of the above'
            ],
            infoText, CbLabel='', parent=self
        )
        win.exec_()
        return win.selectedItemIdx, win.cancel

    def instructMoveRawFiles(self):
        msg = widgets.myMessageBox(showCentered=False, wrapText=False)
        txt = html_utils.paragraph("""
            Put all of the raw microscopy files from the <b>same experiment</b><br> 
            into an <b>empty folder</b> before closing this dialogue.<br><br>

            Note that there should be <b>no other files</b> in this folder.
        """
        )
        msg.information(
            self, 'Microscopy files location', txt, 
            buttonsTexts=('Cancel', 'Done')
        )
        if msg.cancel:
            return False
        else:
            return True

    def checkFileFormat(self, raw_src_path):
        self.moveOtherFiles = False
        self.copyOtherFiles = False
        ls = natsorted(myutils.listdir(raw_src_path))
        files = [
            filename for filename in ls
            if os.path.isfile(os.path.join(raw_src_path, filename))
        ]
        all_ext = [
            os.path.splitext(filename)[1] for filename in ls
            if os.path.isfile(os.path.join(raw_src_path, filename))
        ]
        counter = Counter(all_ext)
        unique_ext = list(counter.keys())
        is_ext_unique = len(unique_ext) == 1
        most_common_ext, _ = counter.most_common(1)[0]
        if not is_ext_unique:
            msg = QMessageBox()
            proceedWithMostCommon = msg.warning(
                self, 'Multiple extensions detected',
                f'The folder {raw_src_path}<br>'
                'contains files with different file extensions '
                f'(extensions detected: {unique_ext})<br><br>'
                f'However, the most common extension is <b>{most_common_ext}</b>, '
                'do you want to proceed with\n'
                f'loading only files with extension <b>{most_common_ext}</b>?',
                msg.Yes | msg.Cancel
            )
            if proceedWithMostCommon == msg.Yes:
                files = [
                    filename for filename in files
                    if os.path.splitext(filename)[1] == most_common_ext
                ]
                otherExt = [ext for ext in unique_ext if ext != most_common_ext]
                files = self.askActionWithOtherFiles(files, otherExt)
            else:
                return []

        if self.rawDataStruct == 0 and len(files) > 1:
            files = self.warnMultipleFiles(files)

        return files

    def askActionWithOtherFiles(self, files, otherExt):
        self.moveOtherFiles = False
        msg = QMessageBox(self)
        msg.setWindowTitle('Action with the other files?')
        txt = (f"""
        <p style="font-size:11px">
            What should I do with the other files (ext: {otherExt})
            in the folder?<br><br>
            <i>NOTE: Only the files with the same basename and position number
            as the raw files will be moved or copied.</i>
        </p>

        """)
        msg.setIcon(msg.Question)
        msg.setText(txt)
        leaveButton = QPushButton(
                'Leave them where they are'
        )
        moveButton = QPushButton(
                'Attempt MOVING to their Position folder'
        )
        copyButton = QPushButton(
                'Attempt COPYING to their Position folder'
        )
        cancelButton = QPushButton(
                'Cancel'
        )
        msg.addButton(leaveButton, msg.YesRole)
        msg.addButton(moveButton, msg.NoRole)
        msg.addButton(copyButton, msg.RejectRole)
        msg.addButton(cancelButton, msg.ApplyRole)
        msg.exec_()
        if msg.clickedButton() == leaveButton:
            self.moveOtherFiles = False
            self.copyOtherFiles = False
            return files
        elif msg.clickedButton() == moveButton:
            self.moveOtherFiles = True
            self.copyOtherFiles = False
            return files
        elif msg.clickedButton() == copyButton:
            self.moveOtherFiles = False
            self.copyOtherFiles = True
            return files
        elif msg.clickedButton() == cancelButton:
            return []


    def warnMultipleFiles(self, files):
        win = apps.QDialogCombobox(
            'Multiple microscopy files detected!', files,
             '<p style="font-size:13px">'
             'You selected "Single microscopy file", '
             'but the <b>folder contains multiple files</b>.<br>'
             '</p>',
             CbLabel='Select which file to load: ', parent=self,
             iconPixmap=QtGui.QPixmap(':warning.svg')
        )
        win.exec_()
        if win.cancel:
            return []
        else:
            files = [win.selectedItemText]

    def attemptSeparateMultiChannel(self, rawFilenames):
        self.chNames = set()
        self.posNums = set()
        stripped_filenames = []
        for file in rawFilenames:
            filename, ext = os.path.splitext(file)
            m_iter = myutils.findalliter(fr'(\d+)_(.+)', filename)
            if len(m_iter) <= 1:
                self.criticalNoFilenamePattern()
                return False
            else:
                m = m_iter[-2]

            try:
                posNum, chName = int(m[0][0]), m[0][1]
                self.chNames.add(chName)
                self.posNums.add(posNum)
                ch_idx = filename.find(f'{posNum}_{chName}')
                stripped_filenames.append(filename[:ch_idx])
            except Exception as e:
                traceback_str = traceback.format_exc()
                self.logger.exception(traceback_str)
                self.criticalNoFilenamePattern(error=traceback.format_exc())
                return False

        basename = myutils.getBasename(stripped_filenames)
        if not basename:
            self.criticalNoFilenamePattern()
            return False

        self.basename = basename

        self.posNums = sorted(list(self.posNums))
        self.chNames = list(self.chNames)
        self.SizeS = len(self.posNums)
        return True


    def criticalNoFilenamePattern(self, error=''):
        txt = (
        """
        <b>Files are named with a non-compatible pattern.</b><br><br>
        In order to automatically generate the required data structure
        from "Multiple files, one for each channel" the filenames must
        be named with a specific pattern:<br><br>
        - basenameN_channelName1, e.g., [ASY015_1_GFP, ASY015_1_mNeon]<br><br>
        where "ASY015" is the basename common to ALL files, "1" is the Position number
        and ["GFP", "mNeon"] are the channel names.<br><br>
        <i>Note that the channel MUST be separated
        from the rest of the name by an underscore "_"</i>
        """
        )
        msg = QMessageBox(self)
        msg.setWindowTitle('Non-compatible pattern')
        msg.setIcon(msg.Critical)
        msg.setText(txt)
        if error:
            msg.setDetailedText(error)
        msg.exec_()

    def criticalBioFormats(self, actionTxt, tracebackFormat, filename):
        msg = QMessageBox(self)
        msg.setIcon(msg.Critical)
        msg.setWindowTitle('Critical error Bio-Formats')
        msg.setDefaultButton(msg.Ok)

        url = 'https://docs.openmicroscopy.org/bio-formats/6.7.0/supported-formats.html'
        seeHere = f'<a href=\"{url}">here</a>'

        _, ext = os.path.splitext(filename)
        txt = (
            f"""
            <p "font-size:11px">
            Error while {actionTxt} with Bio-Formats.<br><br>

            This is most likely because the <b>file format {ext} is not fully supported</b>
            by the Bio-Formats library.<br><br>

            See {seeHere} for details about supported formats.<br><br>

            Try loading file in Fiji and create the data structure manually.<br><br>
            Alternatively, if you are trying to load a video file, you can try
            to open the main GUI and then go to "File --> Open image/video file..."<br><br>
            You were trying to read file: {filename}
            </p>
            """
        )
        msg.setTextFormat(Qt.RichText)
        msg.setTextInteractionFlags(Qt.TextBrowserInteraction)
        msg.setText(txt)
        msg.setDetailedText(tracebackFormat)
        msg.exec_()
        self.close()

    def askConfirmMetadata(
            self, filename, LensNA, DimensionOrder, SizeT, SizeZ, SizeC, SizeS,
            TimeIncrement, TimeIncrementUnit, PhysicalSizeX, PhysicalSizeY,
            PhysicalSizeZ, PhysicalSizeUnit, chNames, emWavelens, ImageName,
            rawFilePath, sampleImgData
        ):
        if self.rawDataStruct == 2:
            filename = self.basename
        self.metadataWin = apps.QDialogMetadataXML(
            title=f'Metadata for {filename}', rawFilename=filename,
            LensNA=LensNA, DimensionOrder=DimensionOrder,
            SizeT=SizeT, SizeZ=SizeZ, SizeC=SizeC, SizeS=SizeS,
            TimeIncrement=TimeIncrement, TimeIncrementUnit=TimeIncrementUnit,
            PhysicalSizeX=PhysicalSizeX, PhysicalSizeY=PhysicalSizeY,
            PhysicalSizeZ=PhysicalSizeZ, PhysicalSizeUnit=PhysicalSizeUnit,
            ImageName=ImageName, chNames=chNames, emWavelens=emWavelens,
            parent=self, rawDataStruct=self.rawDataStruct,
            sampleImgData=sampleImgData, rawFilePath=rawFilePath
        )
        self.metadataWin.exec_()
        self.worker.metadataWin = self.metadataWin
        self.waitCond.wakeAll()

    def askReplacePos(self, pos_path):
        msg = widgets.myMessageBox()
        txt = html_utils.paragraph(
            f'The following folder <b>already exists</b>.<br><br>'
            f'<code>{pos_path}</code><br><br>'
            'Do you want to <b>overwrite</b> all of its content or '
            '<b>add files</b> to it?'
        )
        cancelButton, overwriteButton, addFilesButton = msg.warning(
           self, 'Replace files?', txt,
           buttonsTexts=('Cancel', 'Overwrite', 'Add files')
        )
        if msg.cancel:
            self.worker.cancel = True
        elif overwriteButton == msg.clickedButton:
            self.worker.overwritePos = True
        elif addFilesButton == msg.clickedButton:
            self.worker.addFiles = True
        self.waitCond.wakeAll()

    def closeEvent(self, event):
        self.logger.info('Closing data structure logger...')
        handlers = self.logger.handlers[:]
        for handler in handlers:
            handler.close()
            self.logger.removeHandler(handler)

        if self.buttonToRestore is not None:
            button, color = self.buttonToRestore
            button.setText('0. Attempt "Create data structure" again')
            button.setStyleSheet(
                f'QPushButton {{background-color: {color};}}')
            self.mainWin.setWindowState(Qt.WindowNoState)
            self.mainWin.setWindowState(Qt.WindowActive)
            self.mainWin.raise_()


if __name__ == "__main__":
    print('Launching data structure module...')
    # Handle high resolution displays:
    if hasattr(Qt, 'AA_EnableHighDpiScaling'):
        QApplication.setAttribute(Qt.AA_EnableHighDpiScaling, True)
    if hasattr(Qt, 'AA_UseHighDpiPixmaps'):
        QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps, True)
    # Create the application
    app = QApplication(sys.argv)
    app.setStyle(QStyleFactory.create('Fusion'))
    app.setWindowIcon(QtGui.QIcon(":icon.ico"))
    try:
        win = createDataStructWin(allowExit=True)
        win.show()
        win.setWindowState(Qt.WindowActive)
        win.raise_()
        win.logger.info('Starting main process...')
        win.main()
    except OSError:
        traceback.print_exc()



from stardist.models import StarDist2D, StarDist3D
from csbdeep.utils import normalize
import numpy as np
import os
import skimage.measure

from cellacdc import models

class Model:
    def __init__(
            self, model_name='2D_versatile_fluo',
            load_stardist_3D=False
        ):
        stardist_default_models = models.STARDIST_MODELS
        stardist_path = os.path.dirname(os.path.abspath(__file__))
        T_cell_path = os.path.join(stardist_path, 'model', 'T_cell')
        model_class = StarDist3D if load_stardist_3D else StarDist2D
        if not os.path.exists(T_cell_path):
            model_name = stardist_default_models[0]
        # Initialize model
        if model_name in stardist_default_models:
            self.model = model_class.from_pretrained(model_name)
        else:
            script_path = os.path.abspath(__file__)
            stardist_path = os.path.dirname(script_path)
            model_path = os.path.join(stardist_path, 'model')
            self.model = model_class(
                None, name=model_name, basedir=model_path
            )
        self.load_stardist_3D = load_stardist_3D

    def segment(
            self, image, prob_thresh=0.0, nms_thresh=0.0,
            segment_3D_volume=False
        ):
        # Check on image shape
        is2D = image.ndim == 2
        is3D = image.ndim == 3
        calling_stardist3D_on_2D_data = (
            (is3D and self.load_stardist_3D and not segment_3D_volume)
            or is2D and self.load_stardist_3D
        )
        calling_stardist2D_on_3D_data = (
            is3D and not self.load_stardist_3D and segment_3D_volume
        )
        if calling_stardist3D_on_2D_data:
            print('')
            print('='*30)
            raise ValueError(
                'StarDist3D cannot segment 2D image data. If you are trying to '
                'segment z-slices one by one you need to click "True" at the '
                '"Segment 3D Volume" entry.'
            )
        elif calling_stardist2D_on_3D_data:
            print('')
            print('='*30)
            raise ValueError(
                'StarDist2D cannot segment 3D image data. If you are trying to '
                'segment z-slices one by one you need to click "False" at the '
                '"Segment 3D Volume" entry.'
            )

        # Preprocess image
        prob_thresh = prob_thresh if prob_thresh > 0 else None
        nms_thresh = nms_thresh if nms_thresh > 0 else None
        if not segment_3D_volume and image.ndim == 3:
            labels = np.zeros(image.shape, dtype=np.uint32)
            for i, _img in enumerate(image):
                lab, _ = self.model.predict_instances(
                    normalize(_img),
                    prob_thresh=prob_thresh,
                    nms_thresh=nms_thresh
                )
                labels[i] = lab
            labels = skimage.measure.label(labels>0)
        else:
            labels, _ = self.model.predict_instances(
                normalize(image),
                prob_thresh=prob_thresh,
                nms_thresh=nms_thresh
            )
        return labels.astype(np.uint32)



import os
import pathlib

import numpy as np

import skimage.exposure
import skimage.filters
import skimage.measure

from cellpose import models
from cellacdc.models import CELLPOSE_MODELS
from cellacdc import printl

class Model:
    def __init__(self, model_type='cyto', net_avg=False, gpu=False):
        if model_type not in CELLPOSE_MODELS:
            err_msg = (
                f'"{model_type}" not available. '
                f'Available models are {CELLPOSE_MODELS}'
            )
            raise NameError(err_msg)
        if model_type=='cyto':
            self.model = models.Cellpose(
                gpu=gpu, net_avg=net_avg, model_type=model_type
            )
        else:
            self.model = models.CellposeModel(
                gpu=gpu, net_avg=net_avg, model_type=model_type
            )
        
    def setupLogger(self, logger):
        models.models_logger = logger
    
    def _eval(self, image, **kwargs):
        return self.model.eval(image.astype(np.float32), **kwargs)[0]
    
    def _initialize_image(self, image):
        # See cellpose.gui.io._initialize_images
        if image.ndim > 3:
            # make tiff Z x channels x W x H
            if image.shape[0]<4:
                # tiff is channels x Z x W x H
                image = np.transpose(image, (1,0,2,3))
            elif image.shape[-1]<4:
                # tiff is Z x W x H x channels
                image = np.transpose(image, (0,3,1,2))
            # fill in with blank channels to make 3 channels
            if image.shape[1] < 3:
                shape = image.shape
                shape_to_concat = (shape[0], 3-shape[1], shape[2], shape[3])
                to_concat = np.zeros(shape_to_concat, dtype=np.uint8)
                image = np.concatenate((image, to_concat), axis=1)
            image = np.transpose(image, (0,2,3,1))
        elif image.ndim==3:
            if image.shape[0] < 5:
                image = np.transpose(image, (1,2,0))
            if image.shape[-1] < 3:
                shape = image.shape
                #if parent.autochannelbtn.isChecked():
                #    image = normalize99(image) * 255
                shape_to_concat = (shape[0], shape[1], 3-shape[2])
                to_concat = np.zeros(shape_to_concat,dtype=type(image[0,0,0]))
                image = np.concatenate((image, to_concat), axis=-1)
                image = image[np.newaxis,...]
            elif image.shape[-1]<5 and image.shape[-1]>2:
                image = image[:,:,:3]
                #if parent.autochannelbtn.isChecked():
                #    image = normalize99(image) * 255
                image = image[np.newaxis,...]
        else:
            image = image[np.newaxis,...]    
        
        img_min = image.min() 
        img_max = image.max()
        image = image.astype(np.float32)
        image -= img_min
        if img_max > img_min + 1e-3:
            image /= (img_max - img_min)
        image *= 255
        if image.ndim < 4:
            image = image[:,:,:,np.newaxis]
        return image
    
    def to_rgb_stack(self, first_ch_data, second_ch_data):
        # The 'cyto' model can work with a second channel (e.g., nucleus).
        # However, it needs to be encoded into one of the RGB channels
        # Here we put the first channel in the 'red' channel and the 
        # second channel in the 'green' channel. We then pass
        # `channels = [1,2]` to the segment method
        rgb_stack = np.zeros((*first_ch_data.shape, 3), dtype=first_ch_data.dtype)
        
        R_slice = [slice(None)]*(rgb_stack.ndim)
        R_slice[-1] = 0
        R_slice = tuple(R_slice)
        rgb_stack[R_slice] = first_ch_data

        G_slice = [slice(None)]*(rgb_stack.ndim)
        G_slice[-1] = 1
        G_slice = tuple(G_slice)

        rgb_stack[G_slice] = second_ch_data

        return rgb_stack
        
    def segment(
            self, image,
            diameter=0.0,
            flow_threshold=0.4,
            cellprob_threshold=0.0,
            stitch_threshold=0.0,
            min_size=15,
            anisotropy=0.0,
            normalize=True,
            resample=True,
            segment_3D_volume=False            
        ):
        # Preprocess image
        # image = image/image.max()
        # image = skimage.filters.gaussian(image, sigma=1)
        # image = skimage.exposure.equalize_adapthist(image)

        isRGB = image.shape[-1] == 3 or image.shape[-1] == 4
        isZstack = (image.ndim==3 and not isRGB) or (image.ndim==4)

        if anisotropy == 0 or not isZstack:
            anisotropy = None
        
        do_3D = segment_3D_volume
        if not isZstack:
            stitch_threshold = 0.0
            segment_3D_volume = False
            do_3D = False
        
        if stitch_threshold > 0:
            do_3D = False

        if flow_threshold==0.0 or isZstack:
            flow_threshold = None

        channels = [0,0] if not isRGB else [1,2]

        eval_kwargs = {
            'channels': channels,
            'diameter': diameter,
            'flow_threshold': flow_threshold,
            'cellprob_threshold': cellprob_threshold,
            'stitch_threshold': stitch_threshold,
            'min_size': min_size,
            'normalize': normalize,
            'do_3D': do_3D,
            'anisotropy': anisotropy,
            'resample': resample
        }

        # Run cellpose eval
        if not segment_3D_volume and isZstack:
            labels = np.zeros(image.shape, dtype=np.uint32)
            for i, _img in enumerate(image):
                _img = self._initialize_image(_img)
                lab = self._eval(_img, **eval_kwargs)
                labels[i] = lab
            labels = skimage.measure.label(labels>0)
        else:
            image = self._initialize_image(image)  
            labels = self._eval(image, **eval_kwargs)
        return labels

def url_help():
    return 'https://cellpose.readthedocs.io/en/latest/api.html'



import numpy as np

import skimage.filters

from PyQt5 import QtGui
from PyQt5.QtCore import Qt, pyqtSignal
from PyQt5.QtWidgets import (
    QDialog, QVBoxLayout, QFormLayout, QHBoxLayout, QComboBox, QDoubleSpinBox,
    QSlider, QCheckBox, QPushButton, QLabel, QGroupBox, QGridLayout,
    QWidget
)

from . import widgets, core, myutils

font = QtGui.QFont()
font.setPixelSize(13)

class FilterBaseDialog(QDialog):
    sigClose = pyqtSignal(object)
    sigApplyFilter = pyqtSignal(str)
    sigPreviewToggled = pyqtSignal(bool, str)
    
    def __init__(
            self, layersChannelNames, winTitle, parent=None, 
            currentChannel=None
        ):
        super().__init__(parent)

        self.cancel = True
        self.setFont(font)

        self.setWindowTitle(winTitle)
        self.setWindowFlags(Qt.Tool | Qt.WindowStaysOnTopHint)

        self.channelsWidget = QWidget()
        layout = QHBoxLayout()
        layout.addWidget(QLabel('Channel to filter:'))

        self.channelsComboBox = QComboBox()
        self.channelsComboBox.addItems(layersChannelNames)
        if currentChannel is not None:
            self.channelsComboBox.setCurrentText(currentChannel)
        layout.addWidget(self.channelsComboBox)
        layout.setStretch(1,1)

        self.channelsWidget.setLayout(layout)
    
    def addChannelName(self, newChannelName):
        self.channelsComboBox.addItem(newChannelName)

    def preview_cb(self, checked):
        channelName = self.channelsComboBox.currentText()
        self.sigPreviewToggled.emit(checked, channelName)

    def apply(self):
        channelName = self.channelsComboBox.currentText()
        self.sigApplyFilter.emit(channelName)
    
    def closeEvent(self, event):
        self.sigClose.emit(self)

class gaussBlurDialog(FilterBaseDialog):
    def __init__(self, layersChannelNames, parent=None, **kwargs):
        currentChannel = kwargs['currentChannel']

        super().__init__(
            layersChannelNames, 'Gaussian blur sigma', parent=parent,
            currentChannel=currentChannel        
        )
        
        mainLayout = QVBoxLayout()
        formLayout = QFormLayout()
        buttonsLayout = QHBoxLayout()

        mainLayout.addWidget(self.channelsWidget)

        self.sigmaQDSB = QDoubleSpinBox()
        self.sigmaQDSB.setAlignment(Qt.AlignCenter)
        self.sigmaQDSB.setSingleStep(0.5)
        self.sigmaQDSB.setValue(1.0)
        formLayout.addRow('Gaussian filter sigma:  ', self.sigmaQDSB)
        formLayout.setContentsMargins(0, 10, 0, 10)

        self.sigmaSlider = QSlider(Qt.Horizontal)
        self.sigmaSlider.setMinimum(0)
        self.sigmaSlider.setMaximum(100)
        self.sigmaSlider.setValue(20)
        self.sigma = 1.0
        self.sigmaSlider.setTickPosition(QSlider.TicksBelow)
        self.sigmaSlider.setTickInterval(10)

        self.PreviewCheckBox = QCheckBox("Preview")
        self.PreviewCheckBox.setChecked(True)

        mainLayout.addLayout(formLayout)
        mainLayout.addWidget(self.sigmaSlider)
        mainLayout.addWidget(self.PreviewCheckBox)

        closeButton = widgets.cancelPushButton('Close')

        buttonsLayout.addStretch(1)
        buttonsLayout.addWidget(closeButton)
        buttonsLayout.setContentsMargins(0, 10, 0, 0)

        mainLayout.addLayout(buttonsLayout)

        self.PreviewCheckBox.toggled.connect(self.preview_cb)
        self.sigmaSlider.sliderMoved.connect(self.sigmaSliderMoved)
        self.sigmaQDSB.valueChanged.connect(self.sigmaQDSB_valueChanged)
        self.channelsComboBox.currentTextChanged.connect(self.apply)
        closeButton.clicked.connect(self.close)

        self.setLayout(mainLayout)
    
    def filter(self, image):
        return skimage.filters.gaussian(image, self.sigma)

    def sigmaQDSB_valueChanged(self, val):
        self.sigma = val
        self.sigmaSlider.sliderMoved.disconnect()
        self.sigmaSlider.setSliderPosition(int(val*20))
        self.sigmaSlider.sliderMoved.connect(self.sigmaSliderMoved)
        self.apply()

    def sigmaSliderMoved(self, intVal):
        self.sigma = intVal/20
        self.sigmaQDSB.valueChanged.disconnect()
        self.sigmaQDSB.setValue(self.sigma)
        self.sigmaQDSB.valueChanged.connect(self.sigmaSliderMoved)
        self.apply()
    
    def show(self):
        super().show()
        self.resize(int(self.width()*1.3), self.height())

class diffGaussFilterDialog(FilterBaseDialog):
    def __init__(self, layersChannelNames, parent=None, **kwargs):
        currentChannel = kwargs['currentChannel']

        super().__init__(
            layersChannelNames, 'Gaussian blur sigma', parent=parent,
            currentChannel=currentChannel        
        )

        is3D = kwargs['is3D']

        mainLayout = QVBoxLayout()
        buttonsLayout = QHBoxLayout()

        mainLayout.addWidget(self.channelsWidget)

        firstGroupbox = QGroupBox('First gaussian filter')
        firstLayout = QVBoxLayout()
        self.firstSigmaSliderYX = widgets.sliderWithSpinBox(
            isFloat=True, title='Sigma YX-direction:',
            title_loc='in_line'
        )
        self.firstSigmaSliderYX.setTickPosition(QSlider.TicksBelow)
        self.firstSigmaSliderYX.setSingleStep(0.5)
        self.firstSigmaSliderYX.setTickInterval(10)
        self.firstSigmaSliderZ = widgets.sliderWithSpinBox(
            isFloat=True, title='Sigma Z-direction:  ',
            title_loc='in_line'
        )
        self.firstSigmaSliderZ.setTickPosition(QSlider.TicksBelow)
        self.firstSigmaSliderZ.setSingleStep(0.5)
        self.firstSigmaSliderZ.setTickInterval(10)
        firstLayout.addWidget(self.firstSigmaSliderYX)
        firstLayout.addWidget(self.firstSigmaSliderZ)
        firstGroupbox.setLayout(firstLayout)

        secondGroupbox = QGroupBox('Second gaussian filter')
        secondLayout = QVBoxLayout()
        self.secondSigmaSliderYX = widgets.sliderWithSpinBox(
            isFloat=True, title='Sigma YX-direction:',
            title_loc='in_line'
        )
        self.secondSigmaSliderYX.setTickPosition(QSlider.TicksBelow)
        self.secondSigmaSliderYX.setSingleStep(0.5)
        self.secondSigmaSliderYX.setTickInterval(10)
        self.secondSigmaSliderYX.setValue(1)

        self.secondSigmaSliderZ = widgets.sliderWithSpinBox(
            isFloat=True, title='Sigma Z-direction:  ',
            title_loc='in_line'
        )
        self.secondSigmaSliderZ.setTickPosition(QSlider.TicksBelow)
        self.secondSigmaSliderZ.setSingleStep(0.5)
        self.secondSigmaSliderZ.setTickInterval(10)

        secondLayout.addWidget(self.secondSigmaSliderYX)
        secondLayout.addWidget(self.secondSigmaSliderZ)
        secondGroupbox.setLayout(secondLayout)

        if not is3D:
            self.firstSigmaSliderZ.hide()
            self.secondSigmaSliderZ.hide()

        self.PreviewCheckBox = QCheckBox('Preview filter')
        self.PreviewCheckBox.setChecked(True)

        cancelButton = widgets.cancelPushButton('Cancel')
        buttonsLayout.addStretch(1)
        buttonsLayout.addWidget(cancelButton)

        mainLayout.addWidget(firstGroupbox)
        mainLayout.addSpacing(20)
        mainLayout.addWidget(secondGroupbox)
        mainLayout.addSpacing(20)
        mainLayout.addWidget(self.PreviewCheckBox)
        mainLayout.addSpacing(10)
        mainLayout.addLayout(buttonsLayout)
        mainLayout.addStretch(1)

        self.setLayout(mainLayout)
        self.setFont(font)

        self.firstSigmaSliderYX.sigValueChange.connect(self.apply)
        self.secondSigmaSliderYX.sigValueChange.connect(self.apply)
        if not is3D:
            self.firstSigmaSliderZ.sigValueChange.connect(self.apply)
            self.secondSigmaSliderZ.sigValueChange.connect(self.apply)

        cancelButton.clicked.connect(self.close)
        self.PreviewCheckBox.toggled.connect(self.preview_cb)
    
    def filter(self, img):
        sigmas1, sigmas2 = self.getSigmas()
        sigma1_yx = sigmas1 if isinstance(sigmas1, float) else sigmas1[1]
        sigma2_yx = sigmas2 if isinstance(sigmas2, float) else sigmas2[1]
        if sigma1_yx>0:
            filtered1 = skimage.filters.gaussian(img, sigma=sigmas1)
        else:
            filtered1 = myutils.uint_to_float(img)

        if sigma2_yx>0:
            filtered2 = skimage.filters.gaussian(img, sigma=sigmas2)
        else:
            filtered2 = myutils.uint_to_float(img)

        resultFiltered = filtered1 - filtered2
        return resultFiltered

    def initSpotmaxValues(self, posData):
        self.firstSigmaSliderYX.setValue(0)
        self.firstSigmaSliderZ.setValue(0)
        PhysicalSizeY = posData.PhysicalSizeY
        PhysicalSizeX = posData.PhysicalSizeX
        PhysicalSizeZ = posData.PhysicalSizeZ
        zyx_vox_dim = [PhysicalSizeZ, PhysicalSizeY, PhysicalSizeX]
        wavelen = 510
        NA = 1.4
        yx_resolution_multi = 1
        z_resolution_limit = 1
        _, zyx_resolution_pxl, _ = core.calc_resolution_limited_vol(
            wavelen, NA, yx_resolution_multi, zyx_vox_dim, z_resolution_limit
        )
        self.secondSigmaSliderYX.setValue(zyx_resolution_pxl[1])
        self.secondSigmaSliderZ.setValue(zyx_resolution_pxl[0])

    def getSigmas(self):
        sigma1_yx = self.firstSigmaSliderYX.value()
        sigma1_z = self.firstSigmaSliderZ.value()
        sigma2_yx = self.secondSigmaSliderYX.value()
        sigma2_z = self.secondSigmaSliderZ.value()
        sigmas1 = (sigma1_z, sigma1_yx, sigma1_yx) if sigma1_z>0 else sigma1_yx
        sigmas2 = (sigma2_z, sigma2_yx, sigma2_yx) if sigma2_z>0 else sigma2_yx
        return sigmas1, sigmas2

    def showEvent(self, event):
        self.resize(int(self.width()*1.5), self.height())
        self.firstSigmaSliderYX.setFocus(True)

class edgeDetectionDialog(FilterBaseDialog):
    def __init__(self, layersChannelNames, parent=None, **kwargs):
        currentChannel = kwargs['currentChannel']

        super().__init__(
            layersChannelNames, 'Gaussian blur sigma', parent=parent,
            currentChannel=currentChannel        
        )

        self.keys = layersChannelNames

        mainLayout = QVBoxLayout()
        paramsLayout = QGridLayout()
        buttonsLayout = QHBoxLayout()

        mainLayout.addWidget(self.channelsWidget)

        row = 0
        sigmaQSLabel = QLabel('Blur:')
        paramsLayout.addWidget(sigmaQSLabel, row, 0)
        row += 1
        self.sigmaValLabel = QLabel('1.00')
        paramsLayout.addWidget(self.sigmaValLabel, row, 1)
        self.sigmaSlider = QSlider(Qt.Horizontal)
        self.sigmaSlider.setMinimum(1)
        self.sigmaSlider.setMaximum(100)
        self.sigmaSlider.setValue(20)
        self.sigma = 1.0
        self.sigmaSlider.setTickPosition(QSlider.TicksBelow)
        self.sigmaSlider.setTickInterval(10)
        paramsLayout.addWidget(self.sigmaSlider, row, 0)

        row += 1
        sharpQSLabel = QLabel('Sharpen:')
        # padding: top, left, bottom, right
        sharpQSLabel.setStyleSheet("font-size:13px; padding:5px 0px 0px 0px;")
        paramsLayout.addWidget(sharpQSLabel, row, 0)
        row += 1
        self.sharpValLabel = QLabel('5.00')
        paramsLayout.addWidget(self.sharpValLabel, row, 1)
        self.sharpSlider = QSlider(Qt.Horizontal)
        self.sharpSlider.setMinimum(1)
        self.sharpSlider.setMaximum(100)
        self.sharpSlider.setValue(50)
        self.radius = 5.0
        self.sharpSlider.setTickPosition(QSlider.TicksBelow)
        self.sharpSlider.setTickInterval(10)
        paramsLayout.addWidget(self.sharpSlider, row, 0)

        row += 1
        self.PreviewCheckBox = QCheckBox("Preview")
        self.PreviewCheckBox.setChecked(True)
        paramsLayout.addWidget(self.PreviewCheckBox, row, 0, 1, 2,
                               alignment=Qt.AlignCenter)


        closeButton = widgets.cancelPushButton('Close')

        buttonsLayout.addStretch(1)
        buttonsLayout.addWidget(closeButton)

        paramsLayout.setContentsMargins(0, 10, 0, 0)
        buttonsLayout.setContentsMargins(0, 10, 0, 0)

        mainLayout.addLayout(paramsLayout)
        mainLayout.addLayout(buttonsLayout)

        self.PreviewCheckBox.clicked.connect(self.preview_cb)
        self.sigmaSlider.sliderMoved.connect(self.sigmaSliderMoved)
        self.sharpSlider.sliderMoved.connect(self.sharpSliderMoved)
        self.channelsComboBox.currentTextChanged.connect(self.apply)
        closeButton.clicked.connect(self.close)

        self.setLayout(mainLayout)

    def filter(self, img):
        edge = skimage.filters.sobel(img)
        img = skimage.filters.gaussian(edge, sigma=self.sigma)
        img = img - skimage.filters.gaussian(img, sigma=self.radius)
        return img

    def sigmaSliderMoved(self, intVal):
        self.sigma = intVal/20
        self.sigmaValLabel.setText(f'{self.sigma:.2f}')
        self.apply()

    def sharpSliderMoved(self, intVal):
        self.radius = 10 - intVal/10
        if self.radius < 0.15:
            self.radius = 0.15
        self.sharpValLabel.setText(f'{intVal/10:.2f}')
        self.apply()
    
    def show(self):
        super().show()
        self.resize(int(self.width()*1.3), self.height())

class entropyFilterDialog(FilterBaseDialog):
    def __init__(self, layersChannelNames, parent=None,  **kwargs):
        currentChannel = kwargs['currentChannel']

        super().__init__(
            layersChannelNames, 'Gaussian blur sigma', parent=parent,
            currentChannel=currentChannel        
        )

        mainLayout = QVBoxLayout()
        paramsLayout = QGridLayout()
        buttonsLayout = QHBoxLayout()

        mainLayout.addWidget(self.channelsWidget)

        row = 0
        sigmaQSLabel = QLabel('Radius: ')
        paramsLayout.addWidget(sigmaQSLabel, row, 0)
        row += 1
        self.radiusValLabel = QLabel('10')
        paramsLayout.addWidget(self.radiusValLabel, row, 1)
        self.radiusSlider = QSlider(Qt.Horizontal)
        self.radiusSlider.setMinimum(1)
        self.radiusSlider.setMaximum(100)
        self.radiusSlider.setValue(10)
        self.radiusSlider.setTickPosition(QSlider.TicksBelow)
        self.radiusSlider.setTickInterval(10)
        paramsLayout.addWidget(self.radiusSlider, row, 0)

        row += 1
        self.PreviewCheckBox = QCheckBox("Preview")
        self.PreviewCheckBox.setChecked(True)
        paramsLayout.addWidget(self.PreviewCheckBox, row, 0, 1, 2,
                               alignment=Qt.AlignCenter)

        closeButton = widgets.cancelPushButton('Close')

        buttonsLayout.addStretch(1)
        buttonsLayout.addWidget(closeButton)

        paramsLayout.setContentsMargins(0, 10, 0, 0)
        buttonsLayout.setContentsMargins(0, 10, 0, 0)

        mainLayout.addLayout(paramsLayout)
        mainLayout.addLayout(buttonsLayout)

        self.PreviewCheckBox.clicked.connect(self.preview_cb)
        self.radiusSlider.sliderMoved.connect(self.radiusSliderMoved)
        self.channelsComboBox.currentTextChanged.connect(self.apply)
        closeButton.clicked.connect(self.close)

        self.setLayout(mainLayout)

    def filter(self, img):
        radius = self.radiusSlider.sliderPosition()
        selem = skimage.morphology.disk(radius)
        if img.ndim == 3:
            filtered = np.zeros(img.shape)
            for z, _img in enumerate(img):
                filtered[z] = skimage.filters.rank.entropy(_img, selem)
        else:
            filtered = skimage.filters.rank.entropy(img, selem)
        return filtered

    def radiusSliderMoved(self, intVal):
        self.radiusValLabel.setText(f'{intVal}')
        self.apply()
    
    def show(self):
        super().show()
        self.resize(int(self.width()*1.3), self.height())



"""
This script runs the tracking U-Net
for 2D images.

@author: jroberts / jamesr787
"""

# Modules:
import os
import numpy as np
from PIL import Image
import cv2
from typing import Dict, Any, Tuple, Union

import delta.utilities as utils
import delta.config as cfg
from delta.assets import download_assets
from delta import pipeline


class FakeReader:

    def __init__(self,
                 x,
                 y,
                 channels,
                 timepoints,
                 filename,
                 original_video,
                 starting_frame):
        """
        Initialize experiment reader

        Parameters
        ----------
        x : int
            Size of images along X axis
        y : int
            Size of images along Y axis
        channels : int
            Number of imaging channels
        timepoints : int
            Number of frames
        filename : str
            File or folder name for the experiment
        original_video : np.array
            3D numpy array of original images
        starting_frame : int
            starting frame to track

        Returns
        -------
        None.

        """
        self.x = x
        self.y = y
        self.channels = channels
        self.timepoints = timepoints
        self.filename = filename
        self.original_video = original_video
        self.starting_frame = starting_frame

    def getframes(self,
                  squeeze_dimensions: bool = True,
                  resize: Tuple[int, int] = None,
                  rescale: Tuple[float, float] = None,
                  globalrescale: Tuple[float, float] = None,
                  rotate: float = None,
                  **kwargs
                  ):
        """
        Get frames from experiment.

        Parameters
        ----------
        squeeze_dimensions : bool, optional
            If True, the numpy squeeze function is applied to the output array,
            removing all singleton dimensions.
            The default is True.
        resize : None or tuple/list of 2 ints, optional
            Dimensions to resize the frames. If None, no resizing is performed.
            The default is None.
        rescale : None or tuple/list of 2 int/floats, optional
            Rescale all values in each frame to be within the given range.
            The default is None.
        globalrescale : None or tuple/list of 2 int/floats, optional
            Rescale all values in all frames to be within the given range.
            The default is None.
        rotate : None or float, optional
            Rotation to apply to the image (in degrees).
            The default is None.

        Returns
        -------
        Numpy Array
            Concatenated frames as requested by the different input options.
            If squeeze_dimensions=False, the array is 5-dimensional, with the
            dimensions order being: Position, Time, Channel, Y, X

        """

        dt: Union[str, type] = self.dtype if rescale is None else np.float32
        if resize is None:
            output = np.empty(
                [self.timepoints, self.y, self.x], dtype=dt
            )
        else:
            output = np.empty(
                [self.timepoints, resize[0], resize[1]],
                dtype=dt,
            )

        # Load images:
        for f in range(self.timepoints):

            idx = f + self.starting_frame
            frame = self.original_video[idx].astype(np.uint16)

            # Optionally resize and rescale:
            if rotate is not None:
                frame = utils.imrotate(frame, rotate)
            if resize is not None:
                frame = cv2.resize(frame, resize[::-1])  # cv2 inverts shape
            if rescale is not None:
                frame = utils.rangescale(frame, rescale)
            # Add to output array:
            output[f] = frame

        # Rescale all images:
        if globalrescale is not None:
            output = utils.rangescale(output, globalrescale)

        output = output[np.newaxis, :, np.newaxis, :, :]

        # Return:
        return np.squeeze(output)[0, :, :] if squeeze_dimensions else output


class tracker:

    def __init__(self, **params):
        """
        Initializes Tracker

        Parameters
        ----------
        params : dict
            Dictionary of parameters:
            'model_type' (2D or mothermachine),
            'original_images_path' (path to original images .tif),
            'single mothermachine chamber' (Bool),
            'verbose' (Bool),
            'legacy' (Bool),
            'pickle' (Bool),
            'movie' (Bool).

        Returns
        -------
        None.

        """

        self.params = params

    def __read_tiff(self,
                    path):
        """
        Reads multipage tiff to numpy array.

        Parameters
        ----------
        path : string
            Path to tiff file.

        Returns
        -------
        images : np.array
            Images as 3D numpy array.

        """
        img = Image.open(path)
        images = []
        for i in range(img.n_frames):
            img.seek(i)
            images.append(np.array(img))
        return np.array(images)

    def __load_model_and_presets(self,
                                 model_type):
        """
        Loads Presets for 2D or mothermachine, initializes model for tracking
        and loads model weights.

        Parameters
        ----------
        model_type : string
            Either '2D' or 'mothermachine'.

        Raises
        ------
        ValueError
            Configuration files or models have not yet been downloaded.
            Downloads the configuration files and model weights.

        Returns
        -------
        None.

        """
        # Load Presets
        while True:
            try:
                # Loads Presets images
                cfg.load_config(presets=model_type)

                # Load models:
                self.models = utils.loadmodels()
                "Dictionary of Tensorflow models"

                break

            except ValueError:
                # Downloads model weights and configuration files for 2D and mothermachine
                download_assets(load_models=True,
                                load_sets=False,
                                load_evals=False,
                                config_level='local')

        if self.params['single mothermachine chamber'] and model_type == 'mothermachine':
            self.models.pop('rois')

    def track(self,
              segm_video,
              signals=None,
              export_to: os.PathLike=None):
        """
        Tracks Cells

        Parameters
        ----------
        segm_video : np.array
            3D numpy array of segmentation mask.

        Returns
        -------
        tracked_video : np.array
            3D numpy array of tracked and labeled video.
        """

        # Loads Presets and Initializes Model
        self.__load_model_and_presets(model_type=self.params['model_type'])

        # Original Shape
        original_shape = segm_video[0].shape

        # Get original video and original image size
        original_video = self.__read_tiff(self.params['original_images_path'])
        reference = utils.rangescale(original_video[0], (0, 1))

        # Preprocess Segmentation Video
        seg_stack = []
        for idx in range(len(segm_video)):
            img = segm_video[idx, :, :]
            if not cfg.crop_windows:
                img = cv2.resize(img, cfg.target_size_seg[::-1])
            img_sm = (img > 0.5).astype(np.uint8)
            if cfg.crop_windows:
                img_sm = img_sm[: original_shape[0], : original_shape[1]].astype(np.uint8)
            seg_stack.append(img_sm)
        segm_video = seg_stack

        # Preprocess Original Video
        box = utils.CroppingBox(
            xtl=0, ytl=0,
            xbr=reference.shape[1], ybr=reference.shape[0],
        )
        img_stack = []
        if len(original_video) != len(segm_video):
            starting_frame = len(original_video) - len(segm_video) - 1
        else:
            starting_frame = 0
        for frame in range(len(segm_video)):
            idx = frame + starting_frame
            # Crop and scale:
            i = utils.rangescale(utils.cropbox(original_video[idx], box), rescale=(0, 1))
            # Append i as is to input images stack:
            img_stack.append(i)

        # Get Save Path (File Name is same as Original Images + .format)
        savepath = self.params['original_images_path']
        filename = savepath.replace('.tif', '')

        # Init reader
        xpreader = FakeReader(x=original_shape[1],
                              y=original_shape[0],
                              channels=0,
                              timepoints=len(segm_video),
                              filename=savepath,
                              original_video=original_video,
                              starting_frame=starting_frame
                              )

        # Init Position
        xp = pipeline.Position(position_nb=0,
                               reader=xpreader,
                               models=self.models,
                               drift_correction=False,
                               crop_windows=cfg.crop_windows)

        # Preprocess
        xp.preprocess(rotation_correction=False)

        # Replace img_stack and seg_stack in ROI
        xp.rois[0].img_stack = img_stack
        xp.rois[0].seg_stack = segm_video

        # Track
        xp.track(frames=list(range(len(segm_video))))

        # Label Cells
        xp.features(frames=list(range(len(segm_video))))

        # Get labels
        tracked_video = np.array(xp.rois[0].label_stack, dtype=np.uint8)

        # Save Results
        if self.params['legacy']:
            xp.legacysave(filename + ".mat")
        if self.params['pickle']:
            import pickle
            with open(filename + ".pkl", "wb") as file:
                pickle.dump(self, file)
        if self.params['movie']:
            movie = xp.results_movie(frames=list(range(len(segm_video))))
            utils.vidwrite(movie, filename + ".mp4", verbose=False)

        # Return tracked and labeled video
        return tracked_video



import os
import sys
import subprocess

try:
    import detectron2
except ModuleNotFoundError:
    if sys.platform.startswith("win"):
        from PyQt5.QtWidgets import QMessageBox, QApplication
        from PyQt5.QtCore import QCoreApplication
        from cellacdc.apps import warnVisualCppRequired

        if QCoreApplication.instance() is None:
            app = QApplication(sys.argv)

        win = warnVisualCppRequired(pkg_name='YeastMate')
        win.exec_()

        subprocess.check_call(
            [sys.executable, '-m', 'pip', 'install', 'Cython']
        )
        # subprocess.check_call(
        #     [sys.executable, '-m', 'pip', 'install',
        #     'git+https://github.com/philferriere/cocoapi.git#subdirectory=PythonAPI']
        # )
    subprocess.check_call(
        [sys.executable, '-m', 'pip', 'install',
        'git+https://github.com/facebookresearch/detectron2.git@v0.5']
    )

try:
    import yeastmatedetector
except ModuleNotFoundError:
    from PyQt5.QtWidgets import QMessageBox, QApplication
    from PyQt5.QtCore import QCoreApplication

    if QCoreApplication.instance() is None:
        app = QApplication(sys.argv)

    from cellacdc import myutils
    cancel = myutils.install_package_msg('YeastMate')
    if cancel:
        raise ModuleNotFoundError(
            'User aborted YeastMate installation'
        )

    subprocess.check_call(
        [sys.executable, '-m', 'pip', 'install',
        'git+https://github.com/hoerlteam/YeastMate']
    )
    # YeastMate installs opencv-python which is not functional with PyQt5 on macOS.
    # Uninstall it, and reinstall opencv-python-headless
    subprocess.check_call(
        [sys.executable, '-m', 'pip', 'uninstall', '-y', 'opencv-python']
    )
    subprocess.check_call(
        [sys.executable, '-m', 'pip', 'uninstall', '-y', 'opencv-python-headless']
    )
    subprocess.check_call(
        [sys.executable, '-m', 'pip', 'install', 'opencv-python-headless']
    )



import numpy as np

import skimage.filters

from cellacdc import printl

class Model:
    def __init__(self):
        pass
    
    def _preprocess(self, img, gauss_sigma):
        if gauss_sigma > 0:
            filtered = skimage.filters.gaussian(img, sigma=gauss_sigma)
            return filtered
        else:
            return img

    def _apply_threshold(self, img, threshold_method):
        thresh_val = getattr(skimage.filters, threshold_method)(img)
        printl(img.shape, img.max(), thresh_val)
        return img > thresh_val
    
    def segment(
            self, image, gauss_sigma=1.0, 
            threshold_method='threshold_otsu',
            segment_3D_volume=False 
        ):
        is3D = image.ndim > 2
        if is3D and not segment_3D_volume:
            # Segment slice-by-slice
            thresh = np.zeros(image.shape, dtype=bool)
            for z, img in enumerate(image):
                filtered = self._preprocess(img, gauss_sigma)
                _thresh = self._apply_threshold(filtered, threshold_method)
                thresh[z] = _thresh
        else:
            filtered = self._preprocess(image, gauss_sigma)
            thresh = self._apply_threshold(filtered, threshold_method)
        
        labels = skimage.measure.label(thresh)

        return labels


import os
from functools import partial

import pandas as pd

from .. import exception_handler
from .. import myutils, apps, widgets, html_utils, printl, workers
from ..utils import base

from PyQt5.QtWidgets import QFileDialog

class ApplyTrackingInfoFromTableUtil(base.MainThreadSinglePosUtilBase):
    def __init__(
            self, app, title: str, infoText: str, parent=None,
            callbackOnFinished=None
        ):
        module = myutils.get_module_name(__file__)
        super().__init__(
            app, title, module, infoText, parent
        )

        self.sigClose.connect(self.close)
        self.callbackOnFinished = callbackOnFinished

    @exception_handler
    def run(self, posPath):
        self.logger.info('Reading exisiting segmentation file names...')
        endFilenameSegm = self.selectSegmFileLoadData(posPath)

        if not endFilenameSegm:
            return False
        
        msg = widgets.myMessageBox(showCentered=False, wrapText=False)
        txt = html_utils.paragraph(
            'After clicking "Ok" you will be asked to <b>select the table '
            'file</b> (.csv) containing the <b>tracking information</b>.'
        )
        msg.information(self, 'Instructions', txt)
        if msg.cancel:
            return False
        
        csvPath = QFileDialog.getOpenFileName(
            self, 'Select table with tracking info', posPath,
            "CSV files (*.csv);;All Files (*)"
        )[0]
        if not csvPath:
            return False

        csvName = os.path.basename(csvPath)
        self.logger.info(f'Reading column names in table "{csvName}"...')

        df = pd.read_csv(csvPath, nrows=2)

        win = apps.ApplyTrackTableSelectColumnsDialog(df, parent=self)
        win.exec_()
        if win.cancel:
            return False
        
        columnsInfo = {
            'frameIndexCol': win.frameIndexCol,
            'trackIDsCol': win.trackedIDsCol,
            'maskIDsCol': win.maskIDsCol,
            'xCentroidCol': win.xCentroidCol,
            'yCentroidCol': win.yCentroidCol,
            'parentIDcol': win.parentIDcol,
            'isFirstFrameOne': win.isFirstFrameOne,
            'deleteUntrackedIDs': win.deleteUntrackedIDs
        }
        
        imagesPath = os.path.join(posPath, 'Images')
        segmFilename = [
            f for f in myutils.listdir(imagesPath) 
            if f.endswith(f'{endFilenameSegm}.npz')
        ][0]
        basename = os.path.splitext(segmFilename)[0]
        overWriteButton = widgets.savePushButton(
            'Overwrite existing segmentation file'
        )
        win = apps.filenameDialog(
            basename=f'{basename}_',
            hintText='Insert a <b>filename for the tracked</b> masks file:',
            allowEmpty=False, defaultEntry='tracked', 
            additionalButtons=(overWriteButton, )
        )
        overWriteButton.clicked.connect(partial(self.overWriteClicked, win))
        win.exec_()
        if win.cancel:
            return False

        self.worker = workers.ApplyTrackInfoWorker(
            self, endFilenameSegm, csvPath, win.filename, columnsInfo, posPath
        )
        if self.callbackOnFinished is not None:
            self.worker.signals.finished.connect(self.callbackOnFinished)
        self.runWorker(self.worker)
        return True
        
    def overWriteClicked(self, win):
        win.cancel = False
        win.filename = ''
        win.close()
        



from PyQt5.QtCore import QObject, pyqtSignal, qInstallMessageHandler
import argparse
import configparser
import pprint

class ConfigParser(configparser.ConfigParser):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.optionxform = str
    
    def __repr__(self) -> str:
        string = pprint.pformat(
            {section: dict(self[section]) for section in self.sections()}
        )
        return string

class QtWarningHandler(QObject):
    sigGeometryWarning = pyqtSignal(object)

    def _resizeWarningHandler(self, msg_type, msg_log_context, msg_string):
        if msg_string.find('Unable to set geometry') != -1:
            try:
                self.sigGeometryWarning.emit(msg_string)
            except:
                pass
        elif msg_string:
            print(msg_string)

warningHandler = QtWarningHandler()
qInstallMessageHandler(warningHandler._resizeWarningHandler)

try:
    ap = argparse.ArgumentParser(description='Cell-ACDC parser')
    ap.add_argument(
        '-d', '--debug', action='store_true',
        help=(
            'Used for debugging. Test code with'
            '"from cellacdc.config import parser_args, debug = parser_args["debug"]", '
            'if debug: <debug code here>'
        )
    )

    # Add dummy argument for stupid Jupyter
    # ap.add_argument('-f')

    parser_args, unknown = ap.parse_known_args()
    parser_args = vars(parser_args)
except:
    print('Importing from notebook, ignoring Cell-ACDC argument parser...')
    parser_args = {}
    parser_args['debug'] = False
